import json
from dataclasses import fields, dataclass
from enum import IntEnum

import random_events
import tqdm
from probabilistic_model.probabilistic_circuit.nx.helper import fully_factorized
from probabilistic_model.probabilistic_circuit.nx.probabilistic_circuit import ProbabilisticCircuit
from probabilistic_model.probabilistic_model import ProbabilisticModel
from random_events.interval import SimpleInterval
from random_events.product_algebra import Event, SimpleEvent
from random_events.set import SetElement, Set
from random_events.variable import Symbolic, Continuous, Variable
from sortedcontainers import SortedSet
from sqlalchemy import select
from typing_extensions import Optional, List, Iterator

from ...action_designator import MoveAndPickUpAction, MoveAndPlaceAction
from ....costmaps import OccupancyCostmap, VisibilityCostmap
from ....datastructures.enums import Arms as EArms, Grasp as EGrasp, TaskStatus
from ....datastructures.pose import PoseStamped
from ....datastructures.world import World
from ....designator import ObjectDesignatorDescription, ActionDescription
from ....failures import ObjectUnreachable, PlanFailure
from ....local_transformer import LocalTransformer
# from ....orm.views import PickUpWithContextView, PlaceWithContextView
from ....world_concepts.world_object import Object


class Grasp(IntEnum):
    FRONT = 0
    LEFT = 1
    RIGHT = 2
    TOP = 3


class Arms(IntEnum):
    LEFT = 0
    RIGHT = 1


class ProbabilisticAction:
    """
    Abstract class for performables that have a probabilistic parametrization.
    """

    @dataclass
    class Variables:
        """
        Variables for probabilistic performables.

        This inner class serves the purpose to define the variables that are used in a model and make them easily
        accessible for the user. The user can access the variables by using the dot notation, e.g. `self.variables.x`.

        The members of this class have to be ordered the same way as the variables in the policy.
        The order of the variables in the policy is most likely alphabetical by name.
        """
        ...

    policy: ProbabilisticCircuit
    """
    The policy that is used to determine the parameters.
    """

    variables: Variables
    """
    The variables of this action.
    """

    def __init__(self, policy: Optional[ProbabilisticCircuit] = None):
        self.variables = self.Variables()
        if policy is None:
            policy = self.default_policy()
        self.policy = policy

    def sample_to_action(self, sample: List) -> ActionDescription:
        """
        Convert a sample from the policy to a performable action.

        :param sample: The sample.
        :return: The action.
        """
        raise NotImplementedError

    def all_variables(self) -> SortedSet:
        """
        :return: All variables of the action.
        """
        return SortedSet([getattr(self.variables, field.name) for field in fields(self.variables)
                          if issubclass(field.type, Variable)])

    def default_policy(self) -> ProbabilisticCircuit:
        """
        :return: The default policy for the action.
        """
        means = {v: 0 for v in self.all_variables() if v.is_numeric}
        variances = {v: 1 for v in self.all_variables() if v.is_numeric}
        model = fully_factorized(self.all_variables(), means, variances)
        return model

    def evidence_from_belief_state(self) -> Event:
        """
        :return: the evidence from the belief state that influences the distribution.
        """
        raise NotImplementedError


class MoveAndPickUp( ProbabilisticAction):

    @dataclass
    class Variables:
        arm: Symbolic = Symbolic("arm", Set.from_iterable(Arms))
        grasp: Symbolic = Symbolic("grasp", Set.from_iterable(Grasp))
        relative_x: Continuous = Continuous("relative_x")
        relative_y: Continuous = Continuous("relative_y")

    variables: Variables

    sample_amount: int = 20
    """
    The amount of samples that should be drawn from the policy when iterating over it.
    """

    object_designator: Object
    """
    The object designator that should be picked up.
    """

    arms: List[Arms]
    """
    The arms that can be used for the pick up.
    """

    grasps: List[Grasp]
    """
    The grasps that can be used for the pick up.
    """

    def __init__(self, object_designator: Object, arms: List[Arms], grasps: List[Grasp],
                 policy: Optional[ProbabilisticCircuit] = None):
        ProbabilisticAction.__init__(self, policy)
        self.object_designator = object_designator
        self.arms = arms
        self.grasps = grasps

    def sample_to_action(self, sample: List) -> MoveAndPickUpAction:
        arm, grasp, relative_x, relative_y = sample
        position = [relative_x, relative_y, 0.]
        pose = PoseStamped(position, frame_id=self.object_designator.tf_frame)
        standing_position = LocalTransformer().transform_pose(pose, "map")
        standing_position.position.z = 0
        action = MoveAndPickUpAction(standing_position, self.object_designator, EArms[Arms(int(arm)).name],
                                     EGrasp(int(grasp)), True, 0.03)
        return action

    def events_from_occupancy_and_visibility_costmap(self) -> Event:
        """
        Create events from the occupancy and visibility costmap.

        :return: The events that can be used as evidence for the model.
        """

        # create occupancy and visibility costmap for the target object
        ocm = OccupancyCostmap(distance_to_obstacle=0.4, from_ros=False, size=200, resolution=0.1,
                               origin=self.object_designator.pose)
        vcm = VisibilityCostmap(min_height=1.27, max_height=1.69,
                                size=200, resolution=0.1, origin=self.object_designator.pose)
        mcm = ocm + vcm

        # convert rectangles to events
        events = []
        for rectangle in mcm.partitioning_rectangles():
            event = SimpleEvent({self.variables.relative_x: random_events.interval.open(rectangle.x_lower, rectangle.x_upper),
                                 self.variables.relative_y: random_events.interval.open(rectangle.y_lower, rectangle.y_upper)})
            events.append(event)
        return Event(*events)

    def ground_model(self, model: Optional[ProbabilisticCircuit] = None,
                     event: Optional[Event] = None) -> ProbabilisticCircuit:
        """
        Ground the model to the current evidence.

        :param model: The model that should be grounded. If None, the policy is used.
        :param event: The events that should be used as evidence. If None, the occupancy costmap is used.
        :return: The grounded model
        """

        if model is None:
            model = self.policy
        if event is None:
            event = self.events_from_occupancy_and_visibility_costmap()

        result, probability = model.conditional(event)

        if probability == 0:
            raise ObjectUnreachable("No possible locations found")

        return result

    def iter_with_mode(self) -> Iterator[MoveAndPickUpAction]:
        """
        Generate performables by sampling from the mode of the policy conditioned on visibility and occupancy.
        """
        model = self.ground_model()

        modes, log_max = model.log_mode(check_determinism=False)
        model, _ = model.conditional(modes)
        samples = model.sample(self.sample_amount)

        for sample in samples:
            action = self.sample_to_action(sample)
            yield action

    def __iter__(self) -> Iterator[MoveAndPickUpAction]:
        """
        Generate performables by sampling from the policy conditioned on visibility and occupancy.
        """
        model = self.ground_model(self.policy, self.events_from_occupancy_and_visibility_costmap())
        samples = model.sample(self.sample_amount)
        likelihoods = model.likelihood(samples)

        # sort samples by likelihood
        samples = [x for _, x in sorted(zip(likelihoods, samples), key=lambda pair: pair[0], reverse=True)]

        for sample in samples:
            action = self.sample_to_action(sample)
            yield action

    def iterate_without_occupancy_costmap(self) -> Iterator[MoveAndPickUpAction]:
        """
        Generate performables by sampling from the policy without conditioning on visibility and occupancy.
        """
        samples = self.policy.sample(self.sample_amount)
        for sample in samples:
            action = self.sample_to_action(sample)
            yield action

    @staticmethod
    def query_for_database():
        view = PickUpWithContextView
        query = (select(view.arm, view.grasp, view.relative_x, view.relative_y)
                 .where(view.status == TaskStatus.SUCCEEDED))
        return query

    def batch_rollout(self):
        """
        Try the policy without conditioning on visibility and occupancy and count the successful tries.
        """

        # initialize statistics
        successful_tries = 0
        total_tries = 0

        # create the progress bar
        progress_bar = tqdm.tqdm(iter(self.iterate_without_occupancy_costmap()), total=self.sample_amount)

        for action in progress_bar:
            total_tries += 1
            try:
                action.perform()
                successful_tries += 1
            except PlanFailure as p:
                pass

            # update progress bar
            progress_bar.set_postfix({"Success Probability": successful_tries / total_tries})

            # reset world
            World.current_world.reset_world()


class MoveAndPlace(ProbabilisticAction):

    @dataclass
    class Variables:
        relative_x: Continuous = Continuous("relative_x")
        relative_y: Continuous = Continuous("relative_y")

    variables: Variables

    sample_amount: int = 20
    """
    The amount of samples that should be drawn from the policy when iterating over it.
    """

    object_designator: Object
    """
    The object designator that should be picked up.
    """

    target_location: PoseStamped
    """
    The position to place the object.
    """

    arms: List[Arms]
    """
    The arms that can be used for the pick up.
    """

    def __init__(self, object_designator: Object,
                 target_location: PoseStamped, policy: Optional[ProbabilisticCircuit] = None):
        ProbabilisticAction.__init__(self, policy)
        self.object_designator = object_designator
        self.target_location = target_location

    def sample_to_action(self, sample: List) -> MoveAndPlaceAction:
        relative_x, relative_y = sample
        position = [relative_x + self.target_location.position.x, relative_y + self.target_location.position.y, 0.]
        pose = PoseStamped(position, frame_id=self.target_location.frame_id)
        standing_position = LocalTransformer().transform_pose(pose, "map")
        standing_position.position.z = 0
        action = MoveAndPlaceAction(standing_position, self.object_designator,
                                    self.target_location, EArms.LEFT, True)
        return action

    def events_from_occupancy_and_visibility_costmap(self) -> Event:
        """
        Create events from the occupancy and visibility costmap.

        :return: The events that can be used as evidence for the model.
        """

        # create occupancy and visibility costmap for the target object
        ocm = OccupancyCostmap(distance_to_obstacle=0.4, from_ros=False, size=200, resolution=0.1,
                               origin=self.target_location)
        vcm = VisibilityCostmap(min_height=1.27, max_height=1.69,
                                size=200, resolution=0.1, origin=self.target_location)
        mcm = ocm + vcm

        # convert rectangles to events
        events = []
        for rectangle in mcm.partitioning_rectangles():
            event = SimpleEvent({self.variables.relative_x: random_events.interval.open(rectangle.x_lower, rectangle.x_upper),
                                 self.variables.relative_y: random_events.interval.open(rectangle.y_lower, rectangle.y_upper)})
            events.append(event)
        return Event(*events)

    def ground_model(self, model: Optional[ProbabilisticCircuit] = None,
                     event: Optional[Event] = None) -> ProbabilisticCircuit:
        """
        Ground the model to the current evidence.

        :param model: The model that should be grounded. If None, the policy is used.
        :param event: The events that should be used as evidence. If None, the occupancy costmap is used.
        :return: The grounded model
        """

        if model is None:
            model = self.policy
        if event is None:
            event = self.events_from_occupancy_and_visibility_costmap()

        result, probability = model.conditional(event)

        if probability == 0:
            raise ObjectUnreachable("No possible locations found")

        return result

    def iter_with_mode(self) -> Iterator[MoveAndPlaceAction]:
        """
        Generate performables by sampling from the mode of the policy conditioned on visibility and occupancy.
        """
        model = self.ground_model()
        modes, _ = model.log_mode(check_determinism=False)
        model = self.ground_model(model, modes)
        samples = model.sample(self.sample_amount)

        for sample in samples:
            action = self.sample_to_action(sample)
            yield action

    def __iter__(self) -> Iterator[MoveAndPickUpAction]:
        """
        Generate performables by sampling from the policy conditioned on visibility and occupancy.
        """
        model = self.ground_model(self.policy, self.events_from_occupancy_and_visibility_costmap())
        samples = model.sample(self.sample_amount)
        likelihoods = model.likelihood(samples)

        # sort samples by likelihood
        samples = [x for _, x in sorted(zip(likelihoods, samples), key=lambda pair: pair[0], reverse=True)]

        for sample in samples:
            action = self.sample_to_action(sample)
            yield action

    def iterate_without_occupancy_costmap(self) -> Iterator[MoveAndPickUpAction]:
        """
        Generate performables by sampling from the policy without conditioning on visibility and occupancy.
        """
        samples = self.policy.sample(self.sample_amount)
        for sample in samples:
            action = self.sample_to_action(sample)
            yield action

    @staticmethod
    def query_for_database():
        view = PlaceWithContextView
        query = (select(view.arm, view.relative_x, view.relative_y).where(view.status == TaskStatus.SUCCEEDED))
        return query

    def batch_rollout(self):

        # initialize statistics
        successful_tries = 0
        total_tries = 0

        # create the progress bar
        progress_bar = tqdm.tqdm(iter(self.iterate_without_occupancy_costmap()), total=self.sample_amount)

        for action in progress_bar:
            total_tries += 1
            try:
                action.perform()
                successful_tries += 1
            except PlanFailure as p:
                pass

            # update progress bar
            progress_bar.set_postfix({"Success Probability": successful_tries / total_tries})

            # reset world
            World.current_world.reset_world()



import os.path
import pathlib
from copy import deepcopy

from pycram.testing import BulletWorldTestCase
from pycram.world_concepts.world_object import Object
from pycrap.ontologies import Food


class DescriptionTest(BulletWorldTestCase):
    def test_is_root(self):
        self.assertTrue(self.robot.root_link.is_root)
        self.assertFalse(self.robot.links["base_link"].is_root)

    def test_orientation_as_list(self):
        base_link = self.robot.links["base_link"]
        self.assertEqual(base_link.orientation.to_list(), base_link.pose.orientation.to_list())

    def test_pose_as_list(self):
        base_link = self.robot.links["base_link"]
        self.assertEqual(base_link.pose_as_list, base_link.pose.to_list())

    def test_joint_child_link(self):
        self.assertEqual(self.robot.get_joint_child_link("base_footprint_joint"), self.robot.get_link("base_link"))
        self.assertEqual(self.robot.get_joint_parent_link("base_footprint_joint"), self.robot.root_link)

    def test_generate_description_from_mesh(self):
        file_path = pathlib.Path(__file__).parent.resolve()
        cache_path = self.world.cache_manager.cache_dir
        cache_path = os.path.join(cache_path, f"{self.milk.description.name}.urdf")
        self.milk.description.generate_from_mesh_file(str(file_path) + "/../resources/milk.stl", "milk", cache_path)
        self.assertTrue(self.world.cache_manager.is_cached(f"{self.milk.name}", self.milk.description))

    def test_generate_description_from_description_file(self):
        file_path = pathlib.Path(__file__).parent.resolve()
        file_extension = self.robot.description.get_file_extension()
        pr2_path = str(file_path) + f"/../resources/robots/{self.robot.description.name}{file_extension}"
        cache_path = self.world.cache_manager.cache_dir
        cache_path = os.path.join(cache_path, f"{self.robot.description.name}.urdf")
        self.robot.description.generate_from_description_file(pr2_path, cache_path)
        self.assertTrue(self.world.cache_manager.is_cached(self.robot.name, self.robot.description))

    def test_merge_descriptions(self):
        cereal_2 = Object("cereal_2", Food, self.cereal.path)
        milk_link_name = self.milk.description.get_root()
        milk_joints_num = len(self.milk.description.joints)
        cereal_joints_num = len(cereal_2.description.joints)
        milk_link_num = len(self.milk.description.links)
        cereal_link_num = len(cereal_2.description.links)
        original_cereal_description = deepcopy(cereal_2.description)
        for in_place in [False, True]:
            for new_description_path in [None, "cereal_milk"]:
                description = cereal_2.description.merge_description(self.milk.description, in_place=in_place,
                                                                     new_description_file=new_description_path)
                if in_place:
                    self.assertTrue(description is cereal_2.description)
                    self.assertEqual(len(cereal_2.description.joints), milk_joints_num + cereal_joints_num + 1)
                    self.assertEqual(len(cereal_2.description.links), milk_link_num + cereal_link_num)
                    self.assertTrue(f"{cereal_2.root_link.name}_{self.milk.root_link.name}_joint"
                                    in cereal_2.description.joint_map.keys())
                    self.assertTrue(milk_link_name not in cereal_2.description.link_map.keys())
                    self.assertTrue(f"{self.milk.description.name}_{milk_link_name}" in cereal_2.description.link_map.keys())
                else:
                    self.assertFalse(cereal_2.description is description)
                    self.assertEqual(len(description.joints), milk_joints_num + cereal_joints_num + 1)
                    self.assertEqual(len(description.links), milk_link_num + cereal_link_num)
                    self.assertTrue(f"{cereal_2.root_link.name}_{self.milk.root_link.name}_joint"
                                    in description.joint_map.keys())
                    self.assertTrue(milk_link_name not in description.link_map.keys())
                    self.assertTrue(
                        f"{self.milk.description.name}_{milk_link_name}" in description.link_map.keys())
                if in_place:
                    cereal_2.description = original_cereal_description
        cereal_2.remove()
        self.world.cache_manager.clear_cache()



from ..ros import  get_ros_package_path

from ..robot_description import RobotDescriptionManager, RobotDescription

# Description for turtlebot3_waffle_pi
filename = get_ros_package_path('pycram') + '/resources/robots/' + "turtlebot" + '.urdf'

turtlebot = RobotDescription("turtlebot", "world", "base_link", "base_joint",
                             filename)

# Add to RobotDescriptionManager
rdm = RobotDescriptionManager()
rdm.register_description(turtlebot)



import rospy
from .ros_tools import  is_master_online

# Check is for sphinx autoAPI to be able to work in a CI workflow
if is_master_online():
    rospy.init_node("pycram")

from .data_types import *
from .ros_tools import  *
from .logging import *
from .action_lib import *
from .service import *
from .publisher import *
from .subscriber import *
from .viz_marker_publisher import *



"""
This file was taken from> https://github.com/DLu/tf_transformations/tree/main
Since it is not released for ROS1 we include it directly in PyCRAM

This library is a reimplementation of the tf/transformations.py library.

https://github.com/ros/geometry/blob/noetic-devel/tf/src/tf/transformations.py

Original author: Christoph Gohlke <http://www.lfd.uci.edu/~gohlke/>
Laboratory for Fluorescence Dynamics, University of California, Irvine

Makes use of https://matthew-brett.github.io/transforms3d/ which is also
a reimplementation of the Gohlke's work, but this maintains the API.
"""

# Copyright (c) 2006, Christoph Gohlke
# Copyright (c) 2006-2009, The Regents of the University of California
# Copyright (c) 2021 PickNik Robotics
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the copyright holders nor the names of any
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import math

import numpy

import transforms3d

TRANSLATION_IDENTITY = [0.0, 0.0, 0.0]
ROTATION_IDENTITY = numpy.identity(3, dtype=numpy.float64)
ZOOM_IDENTITY = [1.0, 1.0, 1.0]
SHEAR_IDENTITY = TRANSLATION_IDENTITY


def identity_matrix():
    """
    Return 4x4 identity/unit matrix.

    >>> I = identity_matrix()
    >>> numpy.allclose(I, numpy.dot(I, I))
    True
    >>> numpy.sum(I), numpy.trace(I)
    (4.0, 4.0)
    >>> numpy.allclose(I, numpy.identity(4, dtype=numpy.float64))
    True

    """
    return transforms3d.affines.compose(TRANSLATION_IDENTITY,
                                        ROTATION_IDENTITY,
                                        ZOOM_IDENTITY)


def translation_matrix(direction):
    """
    Return matrix to translate by direction vector.

    >>> v = numpy.random.random(3) - 0.5
    >>> numpy.allclose(v, translation_matrix(v)[:3, 3])
    True

    """
    return transforms3d.affines.compose(direction,
                                        ROTATION_IDENTITY,
                                        ZOOM_IDENTITY)


def translation_from_matrix(matrix):
    """
    Return translation vector from translation matrix.

    >>> v0 = numpy.random.random(3) - 0.5
    >>> v1 = translation_from_matrix(translation_matrix(v0))
    >>> numpy.allclose(v0, v1)
    True

    """
    return transforms3d.affines.decompose(matrix)[0]


def reflection_matrix(point, normal):
    """
    Return matrix to mirror at plane defined by point and normal vector.

    >>> v0 = numpy.random.random(4) - 0.5
    >>> v0[3] = 1.0
    >>> v1 = numpy.random.random(3) - 0.5
    >>> R = reflection_matrix(v0, v1)
    >>> numpy.allclose(2., numpy.trace(R))
    True
    >>> numpy.allclose(v0, numpy.dot(R, v0))
    True
    >>> v2 = v0.copy()
    >>> v2[:3] += v1
    >>> v3 = v0.copy()
    >>> v2[:3] -= v1
    >>> numpy.allclose(v2, numpy.dot(R, v3))
    True

    """
    # Note: point[3] is ignored
    return transforms3d.reflections.rfnorm2aff(normal, point[:3])


def reflection_from_matrix(matrix):
    """
    Return mirror plane point and normal vector from reflection matrix.

    >>> v0 = numpy.random.random(3) - 0.5
    >>> v1 = numpy.random.random(3) - 0.5
    >>> M0 = reflection_matrix(v0, v1)
    >>> point, normal = reflection_from_matrix(M0)
    >>> M1 = reflection_matrix(point, normal)
    >>> is_same_transform(M0, M1)
    True

    """
    normal, point = transforms3d.reflections.aff2rfnorm(matrix)
    return point, normal


def rotation_matrix(angle, direction, point=None):
    """
    Return matrix to rotate about axis defined by point and direction.

    >>> angle = (random.random() - 0.5) * (2*math.pi)
    >>> direc = numpy.random.random(3) - 0.5
    >>> point = numpy.random.random(3) - 0.5
    >>> R0 = rotation_matrix(angle, direc, point)
    >>> R1 = rotation_matrix(angle-2*math.pi, direc, point)
    >>> is_same_transform(R0, R1)
    True
    >>> R0 = rotation_matrix(angle, direc, point)
    >>> R1 = rotation_matrix(-angle, -direc, point)
    >>> is_same_transform(R0, R1)
    True
    >>> I = numpy.identity(4, numpy.float64)
    >>> numpy.allclose(I, rotation_matrix(math.pi*2, direc))
    True
    >>> numpy.allclose(2., numpy.trace(rotation_matrix(math.pi/2,
    ...                                                direc, point)))
    True

    """
    return transforms3d.axangles.axangle2aff(direction, angle, point=point)


def rotation_from_matrix(matrix):
    """
    Return rotation angle and axis from rotation matrix.

    >>> angle = (random.random() - 0.5) * (2*math.pi)
    >>> direc = numpy.random.random(3) - 0.5
    >>> point = numpy.random.random(3) - 0.5
    >>> R0 = rotation_matrix(angle, direc, point)
    >>> angle, direc, point = rotation_from_matrix(R0)
    >>> R1 = rotation_matrix(angle, direc, point)
    >>> is_same_transform(R0, R1)
    True

    """
    direction, angle, point = transforms3d.axangles.aff2axangle(matrix)
    return angle, direction, point


def scale_matrix(factor, origin=None, direction=None):
    """
    Return matrix to scale by factor around origin in direction.

    Use factor -1 for point symmetry.

    >>> v = (numpy.random.rand(4, 5) - 0.5) * 20.0
    >>> v[3] = 1.0
    >>> S = scale_matrix(-1.234)
    >>> numpy.allclose(numpy.dot(S, v)[:3], -1.234*v[:3])
    True
    >>> factor = random.random() * 10 - 5
    >>> origin = numpy.random.random(3) - 0.5
    >>> direct = numpy.random.random(3) - 0.5
    >>> S = scale_matrix(factor, origin)
    >>> S = scale_matrix(factor, origin, direct)

    """
    return transforms3d.zooms.zfdir2aff(factor,
                                        direction=direction,
                                        origin=origin)


def scale_from_matrix(matrix):
    """
    Return scaling factor, origin and direction from scaling matrix.

    >>> factor = random.random() * 10 - 5
    >>> origin = numpy.random.random(3) - 0.5
    >>> direct = numpy.random.random(3) - 0.5
    >>> S0 = scale_matrix(factor, origin)
    >>> factor, origin, direction = scale_from_matrix(S0)
    >>> S1 = scale_matrix(factor, origin, direction)
    >>> is_same_transform(S0, S1)
    True
    >>> S0 = scale_matrix(factor, origin, direct)
    >>> factor, origin, direction = scale_from_matrix(S0)
    >>> S1 = scale_matrix(factor, origin, direction)
    >>> is_same_transform(S0, S1)
    True

    """
    factor, direction, origin = transforms3d.zooms.aff2zfdir(matrix)
    return factor, origin, direction


def projection_matrix(point, normal, direction=None,
                      perspective=None, pseudo=False):
    """
    Return matrix to project onto plane defined by point and normal.

    Using either perspective point, projection direction, or none of both.

    If pseudo is True, perspective projections will preserve relative depth
    such that Perspective = dot(Orthogonal, PseudoPerspective).

    >>> P = projection_matrix((0, 0, 0), (1, 0, 0))
    >>> numpy.allclose(P[1:, 1:], numpy.identity(4)[1:, 1:])
    True
    >>> point = numpy.random.random(3) - 0.5
    >>> normal = numpy.random.random(3) - 0.5
    >>> direct = numpy.random.random(3) - 0.5
    >>> persp = numpy.random.random(3) - 0.5
    >>> P0 = projection_matrix(point, normal)
    >>> P1 = projection_matrix(point, normal, direction=direct)
    >>> P2 = projection_matrix(point, normal, perspective=persp)
    >>> P3 = projection_matrix(point, normal, perspective=persp, pseudo=True)
    >>> is_same_transform(P2, numpy.dot(P0, P3))
    True
    >>> P = projection_matrix((3, 0, 0), (1, 1, 0), (1, 0, 0))
    >>> v0 = (numpy.random.rand(4, 5) - 0.5) * 20.0
    >>> v0[3] = 1.0
    >>> v1 = numpy.dot(P, v0)
    >>> numpy.allclose(v1[1], v0[1])
    True
    >>> numpy.allclose(v1[0], 3.0-v1[1])
    True

    """
    M = numpy.identity(4)
    point = numpy.array(point[:3], dtype=numpy.float64, copy=False)
    normal = unit_vector(normal[:3])
    if perspective is not None:
        # perspective projection
        perspective = numpy.array(perspective[:3], dtype=numpy.float64,
                                  copy=False)
        M[0, 0] = M[1, 1] = M[2, 2] = numpy.dot(perspective-point, normal)
        M[:3, :3] -= numpy.outer(perspective, normal)
        if pseudo:
            # preserve relative depth
            M[:3, :3] -= numpy.outer(normal, normal)
            M[:3, 3] = numpy.dot(point, normal) * (perspective+normal)
        else:
            M[:3, 3] = numpy.dot(point, normal) * perspective
        M[3, :3] = -normal
        M[3, 3] = numpy.dot(perspective, normal)
    elif direction is not None:
        # parallel projection
        direction = numpy.array(direction[:3], dtype=numpy.float64, copy=False)
        scale = numpy.dot(direction, normal)
        M[:3, :3] -= numpy.outer(direction, normal) / scale
        M[:3, 3] = direction * (numpy.dot(point, normal) / scale)
    else:
        # orthogonal projection
        M[:3, :3] -= numpy.outer(normal, normal)
        M[:3, 3] = numpy.dot(point, normal) * normal
    return M


def projection_from_matrix(matrix, pseudo=False):
    """
    Return projection plane and perspective point from projection matrix.

    Return values are same as arguments for projection_matrix function:
    point, normal, direction, perspective, and pseudo.

    >>> point = numpy.random.random(3) - 0.5
    >>> normal = numpy.random.random(3) - 0.5
    >>> direct = numpy.random.random(3) - 0.5
    >>> persp = numpy.random.random(3) - 0.5
    >>> P0 = projection_matrix(point, normal)
    >>> result = projection_from_matrix(P0)
    >>> P1 = projection_matrix(*result)
    >>> is_same_transform(P0, P1)
    True
    >>> P0 = projection_matrix(point, normal, direct)
    >>> result = projection_from_matrix(P0)
    >>> P1 = projection_matrix(*result)
    >>> is_same_transform(P0, P1)
    True
    >>> P0 = projection_matrix(point, normal, perspective=persp, pseudo=False)
    >>> result = projection_from_matrix(P0, pseudo=False)
    >>> P1 = projection_matrix(*result)
    >>> is_same_transform(P0, P1)
    True
    >>> P0 = projection_matrix(point, normal, perspective=persp, pseudo=True)
    >>> result = projection_from_matrix(P0, pseudo=True)
    >>> P1 = projection_matrix(*result)
    >>> is_same_transform(P0, P1)
    True

    """
    M = numpy.array(matrix, dtype=numpy.float64, copy=False)
    M33 = M[:3, :3]
    l, V = numpy.linalg.eig(M)
    i = numpy.where(abs(numpy.real(l) - 1.0) < 1e-8)[0]
    if not pseudo and len(i):
        # point: any eigenvector corresponding to eigenvalue 1
        point = numpy.real(V[:, i[-1]]).squeeze()
        point /= point[3]
        # direction: unit eigenvector corresponding to eigenvalue 0
        l, V = numpy.linalg.eig(M33)
        i = numpy.where(abs(numpy.real(l)) < 1e-8)[0]
        if not len(i):
            raise ValueError('no eigenvector corresponding to eigenvalue 0')
        direction = numpy.real(V[:, i[0]]).squeeze()
        direction /= vector_norm(direction)
        # normal: unit eigenvector of M33.T corresponding to eigenvalue 0
        l, V = numpy.linalg.eig(M33.T)
        i = numpy.where(abs(numpy.real(l)) < 1e-8)[0]
        if len(i):
            # parallel projection
            normal = numpy.real(V[:, i[0]]).squeeze()
            normal /= vector_norm(normal)
            return point, normal, direction, None, False
        else:
            # orthogonal projection, where normal equals direction vector
            return point, direction, None, None, False
    else:
        # perspective projection
        i = numpy.where(abs(numpy.real(l)) > 1e-8)[0]
        if not len(i):
            raise ValueError(
                'no eigenvector not corresponding to eigenvalue 0')
        point = numpy.real(V[:, i[-1]]).squeeze()
        point /= point[3]
        normal = - M[3, :3]
        perspective = M[:3, 3] / numpy.dot(point[:3], normal)
        if pseudo:
            perspective -= normal
        return point, normal, None, perspective, pseudo


def clip_matrix(left, right, bottom, top, near, far, perspective=False):
    """
    Return matrix to obtain normalized device coordinates from frustum.

    The frustum bounds are axis-aligned along x (left, right),
    y (bottom, top) and z (near, far).

    Normalized device coordinates are in range [-1, 1] if coordinates are
    inside the frustum.

    If perspective is True the frustum is a truncated pyramid with the
    perspective point at origin and direction along z axis, otherwise an
    orthographic canonical view volume (a box).

    Homogeneous coordinates transformed by the perspective clip matrix
    need to be dehomogenized (divided by w coordinate).

    >>> frustum = numpy.random.rand(6)
    >>> frustum[1] += frustum[0]
    >>> frustum[3] += frustum[2]
    >>> frustum[5] += frustum[4]
    >>> M = clip_matrix(*frustum, perspective=False)
    >>> numpy.dot(M, [frustum[0], frustum[2], frustum[4], 1.0])
    array([-1., -1., -1.,  1.])
    >>> numpy.dot(M, [frustum[1], frustum[3], frustum[5], 1.0])
    array([ 1.,  1.,  1.,  1.])
    >>> M = clip_matrix(*frustum, perspective=True)
    >>> v = numpy.dot(M, [frustum[0], frustum[2], frustum[4], 1.0])
    >>> v / v[3]
    array([-1., -1., -1.,  1.])
    >>> v = numpy.dot(M, [frustum[1], frustum[3], frustum[4], 1.0])
    >>> v / v[3]
    array([ 1.,  1., -1.,  1.])

    """
    if left >= right or bottom >= top or near >= far:
        raise ValueError('invalid frustrum')
    if perspective:
        if near <= _EPS:
            raise ValueError('invalid frustum: near <= 0')
        t = 2.0 * near
        M = ((-t/(right-left), 0.0, (right+left)/(right-left), 0.0),
             (0.0, -t/(top-bottom), (top+bottom)/(top-bottom), 0.0),
             (0.0, 0.0, -(far+near)/(far-near), t*far/(far-near)),
             (0.0, 0.0, -1.0, 0.0))
    else:
        M = ((2.0/(right-left), 0.0, 0.0, (right+left)/(left-right)),
             (0.0, 2.0/(top-bottom), 0.0, (top+bottom)/(bottom-top)),
             (0.0, 0.0, 2.0/(far-near), (far+near)/(near-far)),
             (0.0, 0.0, 0.0, 1.0))
    return numpy.array(M, dtype=numpy.float64)


def shear_matrix(angle, direction, point, normal):
    """
    Return matrix to shear by angle along direction vector on shear plane.

    The shear plane is defined by a point and normal vector. The direction
    vector must be orthogonal to the plane's normal vector.

    A point P is transformed by the shear matrix into P" such that
    the vector P-P" is parallel to the direction vector and its extent is
    given by the angle of P-P'-P", where P' is the orthogonal projection
    of P onto the shear plane.

    >>> angle = (random.random() - 0.5) * 4*math.pi
    >>> direct = numpy.random.random(3) - 0.5
    >>> point = numpy.random.random(3) - 0.5
    >>> normal = numpy.cross(direct, numpy.random.random(3))
    >>> S = shear_matrix(angle, direct, point, normal)
    >>> numpy.allclose(1.0, numpy.linalg.det(S))
    True

    """
    return transforms3d.shears.sadn2aff(angle, direction, normal, point)


def shear_from_matrix(matrix):
    """
    Return shear angle, direction and plane from shear matrix.

    >>> angle = (random.random() - 0.5) * 4*math.pi
    >>> direct = numpy.random.random(3) - 0.5
    >>> point = numpy.random.random(3) - 0.5
    >>> normal = numpy.cross(direct, numpy.random.random(3))
    >>> S0 = shear_matrix(angle, direct, point, normal)
    >>> angle, direct, point, normal = shear_from_matrix(S0)
    >>> S1 = shear_matrix(angle, direct, point, normal)
    >>> is_same_transform(S0, S1)
    True

    """
    angle, direction, normal, point = transforms3d.shears.aff2sadn(matrix)
    return angle, direction, point, normal


def decompose_matrix(matrix):
    """
    Return sequence of transformations from transformation matrix.

    matrix : array_like
        Non-degenerative homogeneous transformation matrix

    Return tuple of:
        scale : vector of 3 scaling factors
        shear : list of shear factors for x-y, x-z, y-z axes
        angles : list of Euler angles about static x, y, z axes
        translate : translation vector along x, y, z axes
        perspective : perspective partition of matrix

    Raise ValueError if matrix is of wrong type or degenerative.

    >>> T0 = translation_matrix((1, 2, 3))
    >>> scale, shear, angles, trans, persp = decompose_matrix(T0)
    >>> T1 = translation_matrix(trans)
    >>> numpy.allclose(T0, T1)
    True
    >>> S = scale_matrix(0.123)
    >>> scale, shear, angles, trans, persp = decompose_matrix(S)
    >>> scale[0]
    0.123
    >>> R0 = euler_matrix(1, 2, 3)
    >>> scale, shear, angles, trans, persp = decompose_matrix(R0)
    >>> R1 = euler_matrix(*angles)
    >>> numpy.allclose(R0, R1)
    True

    """
    T, R, Z, S = transforms3d.affines.decompose(matrix)
    angles = euler_from_matrix(R)
    # TODO: Fill in the perspective partition of matrix
    return Z, S, angles, T, None


def compose_matrix(scale=None, shear=None, angles=None, translate=None,
                   perspective=None):
    """
    Return transformation matrix from sequence of transformations.

    This is the inverse of the decompose_matrix function.

    Sequence of transformations:
        scale : vector of 3 scaling factors
        shear : list of shear factors for x-y, x-z, y-z axes
        angles : list of Euler angles about static x, y, z axes
        translate : translation vector along x, y, z axes
        perspective : perspective partition of matrix

    >>> scale = numpy.random.random(3) - 0.5
    >>> shear = numpy.random.random(3) - 0.5
    >>> angles = (numpy.random.random(3) - 0.5) * (2*math.pi)
    >>> trans = numpy.random.random(3) - 0.5
    >>> persp = numpy.random.random(4) - 0.5
    >>> M0 = compose_matrix(scale, shear, angles, trans, persp)
    >>> result = decompose_matrix(M0)
    >>> M1 = compose_matrix(*result)
    >>> is_same_transform(M0, M1)
    True

    """
    T = translate if translate is not None else TRANSLATION_IDENTITY
    if angles is not None:
        R = transforms3d.euler.euler2mat(*angles)
    else:
        R = ROTATION_IDENTITY
    Z = scale if scale is not None else ZOOM_IDENTITY
    S = shear if shear is not None else SHEAR_IDENTITY
    M = transforms3d.affines.compose(T, R, Z, S)

    # TODO: Include perspective in composition
    return M


def orthogonalization_matrix(lengths, angles):
    """
    Return orthogonalization matrix for crystallographic cell coordinates.

    Angles are expected in degrees.

    The de-orthogonalization matrix is the inverse.

    >>> O = orthogonalization_matrix((10., 10., 10.), (90., 90., 90.))
    >>> numpy.allclose(O[:3, :3], numpy.identity(3, float) * 10)
    True
    >>> O = orthogonalization_matrix([9.8, 12.0, 15.5], [87.2, 80.7, 69.7])
    >>> numpy.allclose(numpy.sum(O), 43.063229)
    True

    """
    a, b, c = lengths
    angles = numpy.radians(angles)
    sina, sinb, _ = numpy.sin(angles)
    cosa, cosb, cosg = numpy.cos(angles)
    co = (cosa * cosb - cosg) / (sina * sinb)
    return numpy.array((
        (a*sinb*math.sqrt(1.0-co*co),  0.0,    0.0, 0.0),
        (-a*sinb*co,                    b*sina, 0.0, 0.0),
        (a*cosb,                       b*cosa, c,   0.0),
        (0.0,                          0.0,    0.0, 1.0)),
        dtype=numpy.float64)


def superimposition_matrix(v0, v1, scaling=False, usesvd=True):
    """
    Return matrix to transform given vector set into second vector set.

    v0 and v1 are shape (3, *) or (4, *) arrays of at least 3 vectors.

    If usesvd is True, the weighted sum of squared deviations (RMSD) is
    minimized according to the algorithm by W. Kabsch [8]. Otherwise the
    quaternion based algorithm by B. Horn [9] is used (slower when using
    this Python implementation).

    The returned matrix performs rotation, translation and uniform scaling
    (if specified).

    >>> v0 = numpy.random.rand(3, 10)
    >>> M = superimposition_matrix(v0, v0)
    >>> numpy.allclose(M, numpy.identity(4))
    True
    >>> R = random_rotation_matrix(numpy.random.random(3))
    >>> v0 = ((1,0,0), (0,1,0), (0,0,1), (1,1,1))
    >>> v1 = numpy.dot(R, v0)
    >>> M = superimposition_matrix(v0, v1)
    >>> numpy.allclose(v1, numpy.dot(M, v0))
    True
    >>> v0 = (numpy.random.rand(4, 100) - 0.5) * 20.0
    >>> v0[3] = 1.0
    >>> v1 = numpy.dot(R, v0)
    >>> M = superimposition_matrix(v0, v1)
    >>> numpy.allclose(v1, numpy.dot(M, v0))
    True
    >>> S = scale_matrix(random.random())
    >>> T = translation_matrix(numpy.random.random(3)-0.5)
    >>> M = concatenate_matrices(T, R, S)
    >>> v1 = numpy.dot(M, v0)
    >>> v0[:3] += numpy.random.normal(0.0, 1e-9, 300).reshape(3, -1)
    >>> M = superimposition_matrix(v0, v1, scaling=True)
    >>> numpy.allclose(v1, numpy.dot(M, v0))
    True
    >>> M = superimposition_matrix(v0, v1, scaling=True, usesvd=False)
    >>> numpy.allclose(v1, numpy.dot(M, v0))
    True
    >>> v = numpy.empty((4, 100, 3), dtype=numpy.float64)
    >>> v[:, :, 0] = v0
    >>> M = superimposition_matrix(v0, v1, scaling=True, usesvd=False)
    >>> numpy.allclose(v1, numpy.dot(M, v[:, :, 0]))
    True

    """
    v0 = numpy.array(v0, dtype=numpy.float64, copy=False)[:3]
    v1 = numpy.array(v1, dtype=numpy.float64, copy=False)[:3]

    if v0.shape != v1.shape or v0.shape[1] < 3:
        raise ValueError('Vector sets are of wrong shape or type.')

    # move centroids to origin
    t0 = numpy.mean(v0, axis=1)
    t1 = numpy.mean(v1, axis=1)
    v0 = v0 - t0.reshape(3, 1)
    v1 = v1 - t1.reshape(3, 1)

    if usesvd:
        # Singular Value Decomposition of covariance matrix
        u, s, vh = numpy.linalg.svd(numpy.dot(v1, v0.T))
        # rotation matrix from SVD orthonormal bases
        R = numpy.dot(u, vh)
        if numpy.linalg.det(R) < 0.0:
            # R does not constitute right handed system
            R -= numpy.outer(u[:, 2], vh[2, :]*2.0)
            s[-1] *= -1.0
        # homogeneous transformation matrix
        M = numpy.identity(4)
        M[:3, :3] = R
    else:
        # compute symmetric matrix N
        xx, yy, zz = numpy.sum(v0 * v1, axis=1)
        xy, yz, zx = numpy.sum(v0 * numpy.roll(v1, -1, axis=0), axis=1)
        xz, yx, zy = numpy.sum(v0 * numpy.roll(v1, -2, axis=0), axis=1)
        N = ((xx+yy+zz, yz-zy,    zx-xz,    xy-yx),
             (yz-zy,    xx-yy-zz, xy+yx,    zx+xz),
             (zx-xz,    xy+yx,   -xx+yy-zz, yz+zy),
             (xy-yx,    zx+xz,    yz+zy,   -xx-yy+zz))
        # quaternion: eigenvector corresponding to most positive eigenvalue
        l, V = numpy.linalg.eig(N)
        q = V[:, numpy.argmax(l)]
        q /= vector_norm(q)  # unit quaternion
        q = numpy.roll(q, -1)  # move w component to end
        # homogeneous transformation matrix
        M = quaternion_matrix(q)

    # scale: ratio of rms deviations from centroid
    if scaling:
        v0 *= v0
        v1 *= v1
        M[:3, :3] *= math.sqrt(numpy.sum(v1) / numpy.sum(v0))

    # translation
    M[:3, 3] = t1
    T = numpy.identity(4)
    T[:3, 3] = -t0
    M = numpy.dot(M, T)
    return M


def euler_matrix(ai, aj, ak, axes='sxyz'):
    """
    Return homogeneous rotation matrix from Euler angles and axis sequence.

    ai, aj, ak : Euler's roll, pitch and yaw angles
    axes : One of 24 axis sequences as string or encoded tuple

    >>> R = euler_matrix(1, 2, 3, 'syxz')
    >>> numpy.allclose(numpy.sum(R[0]), -1.34786452)
    True
    >>> R = euler_matrix(1, 2, 3, (0, 1, 0, 1))
    >>> numpy.allclose(numpy.sum(R[0]), -0.383436184)
    True
    >>> ai, aj, ak = (4.0*math.pi) * (numpy.random.random(3) - 0.5)
    >>> for axes in _AXES2TUPLE.keys():
    ...    R = euler_matrix(ai, aj, ak, axes)
    >>> for axes in _TUPLE2AXES.keys():
    ...    R = euler_matrix(ai, aj, ak, axes)

    """
    rotation_matrix = transforms3d.euler.euler2mat(ai, aj, ak, axes=axes)
    return transforms3d.affines.compose(TRANSLATION_IDENTITY,
                                        rotation_matrix,
                                        ZOOM_IDENTITY)


def euler_from_matrix(matrix, axes='sxyz'):
    """
    Return Euler angles from rotation matrix for specified axis sequence.

    axes : One of 24 axis sequences as string or encoded tuple

    Note that many Euler angle triplets can describe one matrix.

    >>> R0 = euler_matrix(1, 2, 3, 'syxz')
    >>> al, be, ga = euler_from_matrix(R0, 'syxz')
    >>> R1 = euler_matrix(al, be, ga, 'syxz')
    >>> numpy.allclose(R0, R1)
    True
    >>> angles = (4.0*math.pi) * (numpy.random.random(3) - 0.5)
    >>> for axes in _AXES2TUPLE.keys():
    ...    R0 = euler_matrix(axes=axes, *angles)
    ...    R1 = euler_matrix(axes=axes, *euler_from_matrix(R0, axes))
    ...    if not numpy.allclose(R0, R1): print axes, "failed"

    """
    return transforms3d.euler.mat2euler(matrix, axes=axes)


def euler_from_quaternion(quaternion, axes='sxyz'):
    """
    Return Euler angles from quaternion for specified axis sequence.

    >>> angles = euler_from_quaternion([0.06146124, 0, 0, 0.99810947])
    >>> numpy.allclose(angles, [0.123, 0, 0])
    True

    """
    return euler_from_matrix(quaternion_matrix(quaternion), axes)


def _reorder_input_quaternion(quaternion):
    """Reorder quaternion to have w term first."""
    x, y, z, w = quaternion
    return w, x, y, z


def _reorder_output_quaternion(quaternion):
    """Reorder quaternion to have w term last."""
    w, x, y, z = quaternion
    return x, y, z, w


def quaternion_from_euler(ai, aj, ak, axes='sxyz'):
    """
    Return quaternion from Euler angles and axis sequence.

    ai, aj, ak : Euler's roll, pitch and yaw angles
    axes : One of 24 axis sequences as string or encoded tuple

    >>> q = quaternion_from_euler(1, 2, 3, 'ryxz')
    >>> numpy.allclose(q, [0.310622, -0.718287, 0.444435, 0.435953])
    True

    """
    return _reorder_output_quaternion(
        transforms3d.euler.euler2quat(ai, aj, ak, axes=axes)
    )


def quaternion_about_axis(angle, axis):
    """
    Return quaternion for rotation about axis.

    >>> q = quaternion_about_axis(0.123, (1, 0, 0))
    >>> numpy.allclose(q, [0.06146124, 0, 0, 0.99810947])
    True

    """
    return _reorder_output_quaternion(
        transforms3d.quaternions.axangle2quat(axis, angle)
    )


def quaternion_matrix(quaternion):
    """
    Return 4x4 homogeneous rotation matrix from quaternion.

    >>> R = quaternion_matrix([0.06146124, 0, 0, 0.99810947])
    >>> numpy.allclose(R, rotation_matrix(0.123, (1, 0, 0)))
    True

    """
    rotation_matrix = transforms3d.quaternions.quat2mat(
        _reorder_input_quaternion(quaternion)
    )
    return transforms3d.affines.compose(TRANSLATION_IDENTITY,
                                        rotation_matrix,
                                        ZOOM_IDENTITY)


def quaternion_from_matrix(matrix):
    """
    Return quaternion from rotation matrix.

    >>> R = rotation_matrix(0.123, (1, 2, 3))
    >>> q = quaternion_from_matrix(R)
    >>> numpy.allclose(q, [0.0164262, 0.0328524, 0.0492786, 0.9981095])
    True

    """
    rotation_matrix = transforms3d.affines.decompose(matrix)[1]
    return _reorder_output_quaternion(
        transforms3d.quaternions.mat2quat(rotation_matrix)
    )


def quaternion_multiply(quaternion1, quaternion0):
    """
    Return multiplication of two quaternions.

    >>> q = quaternion_multiply([1, -2, 3, 4], [-5, 6, 7, 8])
    >>> numpy.allclose(q, [-44, -14, 48, 28])
    True

    """
    x0, y0, z0, w0 = quaternion0
    x1, y1, z1, w1 = quaternion1
    w2, x2, y2, z2 = transforms3d.quaternions.qmult([w1, x1, y1, z1],
                                                    [w0, x0, y0, z0])
    return x2, y2, z2, w2


def quaternion_conjugate(quaternion):
    """
    Return conjugate of quaternion.

    >>> q0 = random_quaternion()
    >>> q1 = quaternion_conjugate(q0)
    >>> q1[3] == q0[3] and all(q1[:3] == -q0[:3])
    True

    """
    return numpy.array((-quaternion[0], -quaternion[1],
                        -quaternion[2], quaternion[3]), dtype=numpy.float64)


def quaternion_inverse(quaternion):
    """
    Return inverse of quaternion.

    >>> q0 = random_quaternion()
    >>> q1 = quaternion_inverse(q0)
    >>> numpy.allclose(quaternion_multiply(q0, q1), [0, 0, 0, 1])
    True

    """
    return quaternion_conjugate(quaternion) / numpy.dot(quaternion, quaternion)


def quaternion_slerp(quat0, quat1, fraction, spin=0, shortestpath=True):
    """
    Return spherical linear interpolation between two quaternions.

    >>> q0 = random_quaternion()
    >>> q1 = random_quaternion()
    >>> q = quaternion_slerp(q0, q1, 0.0)
    >>> numpy.allclose(q, q0)
    True
    >>> q = quaternion_slerp(q0, q1, 1.0, 1)
    >>> numpy.allclose(q, q1)
    True
    >>> q = quaternion_slerp(q0, q1, 0.5)
    >>> angle = math.acos(numpy.dot(q0, q))
    >>> numpy.allclose(2.0, math.acos(numpy.dot(q0, q1)) / angle) or
        numpy.allclose(2.0, math.acos(-numpy.dot(q0, q1)) / angle)
    True

    """
    q0 = unit_vector(quat0[:4])
    q1 = unit_vector(quat1[:4])
    if fraction == 0.0:
        return q0
    elif fraction == 1.0:
        return q1
    d = numpy.dot(q0, q1)
    if abs(abs(d) - 1.0) < _EPS:
        return q0
    if shortestpath and d < 0.0:
        # invert rotation
        d = -d
        q1 *= -1.0
    angle = math.acos(d) + spin * math.pi
    if abs(angle) < _EPS:
        return q0
    isin = 1.0 / math.sin(angle)
    q0 *= math.sin((1.0 - fraction) * angle) * isin
    q1 *= math.sin(fraction * angle) * isin
    q0 += q1
    return q0


def random_quaternion(rand=None):
    """
    Return uniform random unit quaternion.

    rand: array like or None
        Three independent random variables that are uniformly distributed
        between 0 and 1.

    >>> q = random_quaternion()
    >>> numpy.allclose(1.0, vector_norm(q))
    True
    >>> q = random_quaternion(numpy.random.random(3))
    >>> q.shape
    (4,)

    """
    if rand is None:
        rand = numpy.random.rand(3)
    else:
        assert len(rand) == 3
    r1 = numpy.sqrt(1.0 - rand[0])
    r2 = numpy.sqrt(rand[0])
    pi2 = math.pi * 2.0
    t1 = pi2 * rand[1]
    t2 = pi2 * rand[2]
    return numpy.array((numpy.sin(t1)*r1,
                        numpy.cos(t1)*r1,
                        numpy.sin(t2)*r2,
                        numpy.cos(t2)*r2), dtype=numpy.float64)


def random_rotation_matrix(rand=None):
    """
    Return uniform random rotation matrix.

    rnd: array like
        Three independent random variables that are uniformly distributed
        between 0 and 1 for each returned quaternion.

    >>> R = random_rotation_matrix()
    >>> numpy.allclose(numpy.dot(R.T, R), numpy.identity(4))
    True

    """
    return quaternion_matrix(random_quaternion(rand))


class Arcball(object):
    """
    Virtual Trackball Control.

    >>> ball = Arcball()
    >>> ball = Arcball(initial=numpy.identity(4))
    >>> ball.place([320, 320], 320)
    >>> ball.down([500, 250])
    >>> ball.drag([475, 275])
    >>> R = ball.matrix()
    >>> numpy.allclose(numpy.sum(R), 3.90583455)
    True
    >>> ball = Arcball(initial=[0, 0, 0, 1])
    >>> ball.place([320, 320], 320)
    >>> ball.setaxes([1,1,0], [-1, 1, 0])
    >>> ball.setconstrain(True)
    >>> ball.down([400, 200])
    >>> ball.drag([200, 400])
    >>> R = ball.matrix()
    >>> numpy.allclose(numpy.sum(R), 0.2055924)
    True
    >>> ball.next()

    """

    def __init__(self, initial=None):
        """
        Initialize virtual trackball control.

        initial : quaternion or rotation matrix

        """
        self._axis = None
        self._axes = None
        self._radius = 1.0
        self._center = [0.0, 0.0]
        self._vdown = numpy.array([0, 0, 1], dtype=numpy.float64)
        self._constrain = False

        if initial is None:
            self._qdown = numpy.array([0, 0, 0, 1], dtype=numpy.float64)
        else:
            initial = numpy.array(initial, dtype=numpy.float64)
            if initial.shape == (4, 4):
                self._qdown = quaternion_from_matrix(initial)
            elif initial.shape == (4, ):
                initial /= vector_norm(initial)
                self._qdown = initial
            else:
                raise ValueError('initial not a quaternion or matrix.')

        self._qnow = self._qpre = self._qdown

    def place(self, center, radius):
        """
        Place Arcball, e.g. when window size changes.

        center : sequence[2]
            Window coordinates of trackball center.
        radius : float
            Radius of trackball in window coordinates.

        """
        self._radius = float(radius)
        self._center[0] = center[0]
        self._center[1] = center[1]

    def setaxes(self, *axes):
        """Set axes to constrain rotations."""
        if axes is None:
            self._axes = None
        else:
            self._axes = [unit_vector(axis) for axis in axes]

    def setconstrain(self, constrain):
        """Set state of constrain to axis mode."""
        self._constrain = constrain is True

    def getconstrain(self):
        """Return state of constrain to axis mode."""
        return self._constrain

    def down(self, point):
        """Set initial cursor window coordinates and pick constrain-axis."""
        self._vdown = arcball_map_to_sphere(point, self._center, self._radius)
        self._qdown = self._qpre = self._qnow

        if self._constrain and self._axes is not None:
            self._axis = arcball_nearest_axis(self._vdown, self._axes)
            self._vdown = arcball_constrain_to_axis(self._vdown, self._axis)
        else:
            self._axis = None

    def drag(self, point):
        """Update current cursor window coordinates."""
        vnow = arcball_map_to_sphere(point, self._center, self._radius)

        if self._axis is not None:
            vnow = arcball_constrain_to_axis(vnow, self._axis)

        self._qpre = self._qnow

        t = numpy.cross(self._vdown, vnow)
        if numpy.dot(t, t) < _EPS:
            self._qnow = self._qdown
        else:
            q = [t[0], t[1], t[2], numpy.dot(self._vdown, vnow)]
            self._qnow = quaternion_multiply(q, self._qdown)

    def next(self, acceleration=0.0):  # noqa: A003 (for backwards compat.)
        """Continue rotation in direction of last drag."""
        q = quaternion_slerp(self._qpre, self._qnow, 2.0+acceleration, False)
        self._qpre, self._qnow = self._qnow, q

    def matrix(self):
        """Return homogeneous rotation matrix."""
        return quaternion_matrix(self._qnow)


def arcball_map_to_sphere(point, center, radius):
    """Return unit sphere coordinates from window coordinates."""
    v = numpy.array(((point[0] - center[0]) / radius,
                     (center[1] - point[1]) / radius,
                     0.0), dtype=numpy.float64)
    n = v[0]*v[0] + v[1]*v[1]
    if n > 1.0:
        v /= math.sqrt(n)  # position outside of sphere
    else:
        v[2] = math.sqrt(1.0 - n)
    return v


def arcball_constrain_to_axis(point, axis):
    """Return sphere point perpendicular to axis."""
    v = numpy.array(point, dtype=numpy.float64, copy=True)
    a = numpy.array(axis, dtype=numpy.float64, copy=True)
    v -= a * numpy.dot(a, v)  # on plane
    n = vector_norm(v)
    if n > _EPS:
        if v[2] < 0.0:
            v *= -1.0
        v /= n
        return v
    if a[2] == 1.0:
        return numpy.array([1, 0, 0], dtype=numpy.float64)
    return unit_vector([-a[1], a[0], 0])


def arcball_nearest_axis(point, axes):
    """Return axis, which arc is nearest to point."""
    point = numpy.array(point, dtype=numpy.float64, copy=False)
    nearest = None
    mx = -1.0
    for axis in axes:
        t = numpy.dot(arcball_constrain_to_axis(point, axis), point)
        if t > mx:
            nearest = axis
            mx = t
    return nearest


# epsilon for testing whether a number is close to zero
_EPS = numpy.finfo(float).eps * 4.0

# axis sequences for Euler angles
_NEXT_AXIS = [1, 2, 0, 1]

# map axes strings to/from tuples of inner axis, parity, repetition, frame
_AXES2TUPLE = {
    'sxyz': (0, 0, 0, 0), 'sxyx': (0, 0, 1, 0), 'sxzy': (0, 1, 0, 0),
    'sxzx': (0, 1, 1, 0), 'syzx': (1, 0, 0, 0), 'syzy': (1, 0, 1, 0),
    'syxz': (1, 1, 0, 0), 'syxy': (1, 1, 1, 0), 'szxy': (2, 0, 0, 0),
    'szxz': (2, 0, 1, 0), 'szyx': (2, 1, 0, 0), 'szyz': (2, 1, 1, 0),
    'rzyx': (0, 0, 0, 1), 'rxyx': (0, 0, 1, 1), 'ryzx': (0, 1, 0, 1),
    'rxzx': (0, 1, 1, 1), 'rxzy': (1, 0, 0, 1), 'ryzy': (1, 0, 1, 1),
    'rzxy': (1, 1, 0, 1), 'ryxy': (1, 1, 1, 1), 'ryxz': (2, 0, 0, 1),
    'rzxz': (2, 0, 1, 1), 'rxyz': (2, 1, 0, 1), 'rzyz': (2, 1, 1, 1)}

_TUPLE2AXES = {v: k for k, v in _AXES2TUPLE.items()}


def vector_norm(data, axis=None, out=None):
    """
    Return length, i.e. eucledian norm, of ndarray along axis.

    >>> v = numpy.random.random(3)
    >>> n = vector_norm(v)
    >>> numpy.allclose(n, numpy.linalg.norm(v))
    True
    >>> v = numpy.random.rand(6, 5, 3)
    >>> n = vector_norm(v, axis=-1)
    >>> numpy.allclose(n, numpy.sqrt(numpy.sum(v*v, axis=2)))
    True
    >>> n = vector_norm(v, axis=1)
    >>> numpy.allclose(n, numpy.sqrt(numpy.sum(v*v, axis=1)))
    True
    >>> v = numpy.random.rand(5, 4, 3)
    >>> n = numpy.empty((5, 3), dtype=numpy.float64)
    >>> vector_norm(v, axis=1, out=n)
    >>> numpy.allclose(n, numpy.sqrt(numpy.sum(v*v, axis=1)))
    True
    >>> vector_norm([])
    0.0
    >>> vector_norm([1.0])
    1.0

    """
    data = numpy.array(data, dtype=numpy.float64, copy=True)
    if out is None:
        if data.ndim == 1:
            return math.sqrt(numpy.dot(data, data))
        data *= data
        out = numpy.atleast_1d(numpy.sum(data, axis=axis))
        numpy.sqrt(out, out)
        return out
    else:
        data *= data
        numpy.sum(data, axis=axis, out=out)
        numpy.sqrt(out, out)


def unit_vector(data, axis=None, out=None):
    """
    Return ndarray normalized by length, i.e. eucledian norm, along axis.

    >>> v0 = numpy.random.random(3)
    >>> v1 = unit_vector(v0)
    >>> numpy.allclose(v1, v0 / numpy.linalg.norm(v0))
    True
    >>> v0 = numpy.random.rand(5, 4, 3)
    >>> v1 = unit_vector(v0, axis=-1)
    >>> v2 = v0 / numpy.expand_dims(numpy.sqrt(numpy.sum(v0*v0, axis=2)), 2)
    >>> numpy.allclose(v1, v2)
    True
    >>> v1 = unit_vector(v0, axis=1)
    >>> v2 = v0 / numpy.expand_dims(numpy.sqrt(numpy.sum(v0*v0, axis=1)), 1)
    >>> numpy.allclose(v1, v2)
    True
    >>> v1 = numpy.empty((5, 4, 3), dtype=numpy.float64)
    >>> unit_vector(v0, axis=1, out=v1)
    >>> numpy.allclose(v1, v2)
    True
    >>> list(unit_vector([]))
    []
    >>> list(unit_vector([1.0]))
    [1.0]

    """
    if out is None:
        data = numpy.array(data, dtype=numpy.float64, copy=True)
        if data.ndim == 1:
            data /= math.sqrt(numpy.dot(data, data))
            return data
    else:
        if out is not data:
            out[:] = numpy.array(data, copy=False)
        data = out
    length = numpy.atleast_1d(numpy.sum(data*data, axis))
    numpy.sqrt(length, length)
    if axis is not None:
        length = numpy.expand_dims(length, axis)
    data /= length
    if out is None:
        return data


def random_vector(size):
    """
    Return array of random doubles in the half-open interval [0.0, 1.0).

    >>> v = random_vector(10000)
    >>> numpy.all(v >= 0.0) and numpy.all(v < 1.0)
    True
    >>> v0 = random_vector(10)
    >>> v1 = random_vector(10)
    >>> numpy.any(v0 == v1)
    False

    """
    return numpy.random.random(size)


def inverse_matrix(matrix):
    """
    Return inverse of square transformation matrix.

    >>> M0 = random_rotation_matrix()
    >>> M1 = inverse_matrix(M0.T)
    >>> numpy.allclose(M1, numpy.linalg.inv(M0.T))
    True
    >>> for size in range(1, 7):
    ...     M0 = numpy.random.rand(size, size)
    ...     M1 = inverse_matrix(M0)
    ...     if not numpy.allclose(M1, numpy.linalg.inv(M0)): print size

    """
    return numpy.linalg.inv(matrix)


def concatenate_matrices(*matrices):
    """
    Return concatenation of series of transformation matrices.

    >>> M = numpy.random.rand(16).reshape((4, 4)) - 0.5
    >>> numpy.allclose(M, concatenate_matrices(M))
    True
    >>> numpy.allclose(numpy.dot(M, M.T), concatenate_matrices(M, M.T))
    True

    """
    M = numpy.identity(4)
    for i in matrices:
        M = numpy.dot(M, i)
    return M


def is_same_transform(matrix0, matrix1):
    """
    Return True if two matrices perform same transformation.

    >>> is_same_transform(numpy.identity(4), numpy.identity(4))
    True
    >>> is_same_transform(numpy.identity(4), random_rotation_matrix())
    False

    """
    matrix0 = numpy.array(matrix0, dtype=numpy.float64, copy=True)
    matrix0 /= matrix0[3, 3]
    matrix1 = numpy.array(matrix1, dtype=numpy.float64, copy=True)
    matrix1 /= matrix1[3, 3]
    return numpy.allclose(matrix0, matrix1)



"""Implementation of helper functions and classes for internal usage only.

Functions:
_block -- wrap multiple statements into a single block.

Classes:
GeneratorList -- implementation of generator list wrappers.
"""
from __future__ import annotations
from inspect import isgeneratorfunction
import os
import math

import numpy as np
from matplotlib import pyplot as plt
import matplotlib.colors as mcolors

from .tf_transformations import quaternion_about_axis, quaternion_multiply
from typing_extensions import Tuple, Callable, List, Dict, TYPE_CHECKING, Sequence, Any, Iterable, Optional

from .datastructures.pose import PoseStamped
from .local_transformer import LocalTransformer

if TYPE_CHECKING:
    from .world_concepts.world_object import Object
    from .robot_description import CameraDescription


def get_rays_from_min_max(min_bound: Sequence[float], max_bound: Sequence[float], step_size_in_meters: float = 0.01) \
        -> np.ndarray:
    """
    Get rays from min and max bounds as an array of start and end 3D points.
    Note: The rays are not steped in the x direction as the rays are cast parallel to the x-axis.

    Example:
    >>> min_bound = [0, 0, 0]
    >>> max_bound = [1, 2, 3]
    >>> rays = get_rays_from_min_max(min_bound, max_bound, 1)
    >>> rays.shape
    (6, 3, 2)
    >>> rays
    array([
    [[0. , 1. ],
     [0. , 0. ],
     [0. , 0. ]],
    [[0. , 1. ],
     [0. , 0. ],
     [1.5, 1.5]],
    [[0. , 1. ],
     [0. , 0. ],
     [3. , 3. ]],
    [[0. , 1. ],
     [2. , 2. ],
     [0. , 0. ]],
    [[0. , 1. ],
     [2. , 2. ],
     [1.5, 1.5]],
    [[0. , 1. ],
     [2. , 2. ],
     [3. , 3. ]]
     ])

    :param min_bound: The minimum bound of the rays, a sequence of 3 floats.
    :param max_bound: The maximum bound of the rays, a sequence of 3 floats.
    :param step_size_in_meters: The step size in meters between the rays.
    :return: The rays as an array of shape (n, 3, 2) where n is number of rays, 3 is because each point has x, y, and z,
    and 2 is for the start and end points of the rays.
    """
    min_bound = np.array(min_bound)
    max_bound = np.array(max_bound)
    n_steps = np.ceil(np.abs(max_bound[1:] - min_bound[1:]) / step_size_in_meters).astype(int)
    rays_start_x = np.ones((n_steps[0], n_steps[1])) * min_bound[0]
    rays_end_x = np.ones((n_steps[0], n_steps[1])) * max_bound[0]
    y_values = np.linspace(min_bound[1], max_bound[1], n_steps[0])
    z_values = np.linspace(min_bound[2], max_bound[2], n_steps[1])
    rays_start_y = np.tile(y_values, (n_steps[1], 1)).T
    rays_end_y = rays_start_y
    rays_start_z = np.tile(z_values, (n_steps[0], 1))
    rays_end_z = rays_start_z
    rays_start = np.stack((rays_start_x, rays_start_y, rays_start_z), axis=-1)
    rays_end = np.stack((rays_end_x, rays_end_y, rays_end_z), axis=-1)
    rays_start = rays_start.reshape(-1, 3)
    rays_end = rays_end.reshape(-1, 3)
    # The shape of rays is (num_rays, 3, 2), while num_rays = n_steps[0] (num y step) * n_steps[1] (num z step)
    return np.stack((rays_start, rays_end), axis=-1)


def chunks(lst: List, n: int) -> List:
    """
    Yield successive n-sized chunks from lst.

    :param lst: The list from which chunks should be yielded
    :param n: Size of the chunks
    :return: A list of size n from lst
    """
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


class bcolors:
    """
    Color codes which can be used to highlight Text in the Terminal. For example,
    for warnings.
    Usage:
    Firstly import the class into the file.
    print(f'{bcolors.WARNING} Some Text {bcolors.ENDC}')
    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def _apply_ik(robot: 'Object', pose_and_joint_poses: Tuple[PoseStamped, Dict[str, float]]) -> None:
    """
    Apllies a list of joint poses calculated by an inverse kinematics solver to a robot

    :param robot: The robot the joint poses should be applied on
    :param pose_and_joint_poses: The base pose and joint states as returned by the ik solver
    :return: None
    """
    pose, joint_states = pose_and_joint_poses
    robot.set_pose(pose)
    robot.set_multiple_joint_positions(joint_states)


class GeneratorList:
    """Implementation of generator list wrappers.

    Generator lists store the elements of a generator, so these can be fetched multiple times.

    Methods:
    get -- get the element at a specific index.
    has -- check if an element at a specific index exists.
    """

    def __init__(self, generator: Callable):
        """Create a new generator list.

        Arguments:
        generator -- the generator to use.
        """
        if isgeneratorfunction(generator):
            self._generator = generator()
        else:
            self._generator = generator

        self._generated = []

    def get(self, index: int = 0):
        """Get the element at a specific index or raise StopIteration if it doesn't exist.

        Arguments:
        index -- the index to get the element of.
        """
        while len(self._generated) <= index:
            self._generated.append(next(self._generator))

        return self._generated[index]

    def has(self, index: int) -> bool:
        """Check if an element at a specific index exists and return True or False.

        Arguments:
        index -- the index to check for.
        """
        try:
            self.get(index)
            return True
        except StopIteration:
            return False


def axis_angle_to_quaternion(axis: List, angle: float) -> Tuple:
    """
    Convert axis-angle to quaternion.

    :param axis: (x, y, z) tuple representing rotation axis.
    :param angle: rotation angle in degree
    :return: The quaternion representing the axis angle
    """
    angle = math.radians(angle)
    axis_length = math.sqrt(sum([i ** 2 for i in axis]))
    normalized_axis = tuple(i / axis_length for i in axis)

    x = normalized_axis[0] * math.sin(angle / 2)
    y = normalized_axis[1] * math.sin(angle / 2)
    z = normalized_axis[2] * math.sin(angle / 2)
    w = math.cos(angle / 2)

    return tuple((x, y, z, w))


class suppress_stdout_stderr(object):
    """
    A context manager for doing a "deep suppression" of stdout and stderr in
    Python, i.e. will suppress all prints, even if the print originates in a
    compiled C/Fortran sub-function.

    This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).
    Copied from https://stackoverflow.com/questions/11130156/suppress-stdout-stderr-print-from-python-functions
    """

    def __init__(self):
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for _ in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        # This one is not needed for URDF parsing output
        # os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        # This one is not needed for URDF parsing output
        # os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)
        # Close all file descriptors
        for fd in self.null_fds + self.save_fds:
            os.close(fd)


def adjust_camera_pose_based_on_target(cam_pose: PoseStamped, target_pose: PoseStamped,
                                       camera_description: CameraDescription) -> PoseStamped:
    """
    Adjust the given cam_pose orientation such that it is facing the target_pose, which partly depends on the
     front_facing_axis of the that is defined in the camera_description.

    :param cam_pose: The camera pose.
    :param target_pose: The target pose.
    :param camera_description: The camera description.
    :return: The adjusted camera pose.
    """
    quaternion = get_quaternion_between_camera_and_target(cam_pose, target_pose, camera_description)
    # apply the rotation to the camera pose using quaternion multiplication
    return apply_quaternion_to_pose(cam_pose, quaternion)


def get_quaternion_between_camera_and_target(cam_pose: PoseStamped, target_pose: PoseStamped,
                                             camera_description: 'CameraDescription') -> np.ndarray:
    """
    Get the quaternion between the camera and the target.

    :param cam_pose: The camera pose.
    :param target_pose: The target pose.
    :param camera_description: The camera description.
    :return: The quaternion between the camera and the target.
    """
    # Get the front facing axis of the camera in the world frame
    front_facing_axis = transform_vector_using_pose(camera_description.front_facing_axis, cam_pose)
    front_facing_axis = front_facing_axis - np.array(cam_pose.position.to_list())

    # Get the vector from the camera to the target
    camera_to_target = cam_pose.get_vector_to_pose(target_pose)

    # Get the quaternion between the camera and target
    return get_quaternion_between_two_vectors(front_facing_axis, camera_to_target)


def transform_vector_using_pose(vector: Sequence, pose) -> np.ndarray:
    """
    Transform a vector using a pose.

    :param vector: The vector.
    :param pose: The pose.
    :return: The transformed vector.
    """
    vector = np.array(vector).reshape(1, 3)
    return pose.to_transform_stamped("pose").apply_transform_to_array_of_points(vector).flatten()


def apply_quaternion_to_pose(pose: PoseStamped, quaternion: np.ndarray) -> PoseStamped:
    """
    Apply a quaternion to a pose.

    :param pose: The pose.
    :param quaternion: The quaternion.
    :return: The new pose.
    """
    pose_quaternion = np.array(pose.orientation.to_list())
    new_quaternion = quaternion_multiply(quaternion, pose_quaternion)
    return PoseStamped(pose.position.to_list(), new_quaternion.tolist())


def get_quaternion_between_two_vectors(v1: np.ndarray, v2: np.ndarray) -> np.ndarray:
    """
    Get the quaternion between two vectors.

    :param v1: The first vector.
    :param v2: The second vector.
    :return: The quaternion between the two vectors.
    """
    axis, angle = get_axis_angle_between_two_vectors(v1, v2)
    return quaternion_about_axis(angle, axis)


def get_axis_angle_between_two_vectors(v1: np.ndarray, v2: np.ndarray) -> Tuple[np.ndarray, float]:
    """
    Get the axis and angle between two vectors.

    :param v1: The first vector.
    :param v2: The second vector.
    :return: The axis and angle between the two vectors.
    """
    v1 = v1 / np.linalg.norm(v1)
    v2 = v2 / np.linalg.norm(v2)
    axis = np.cross(v1, v2)
    angle = np.arccos(np.dot(v1, v2) - 1e-9)  # to avoid numerical errors
    return axis, angle


class RayTestUtils:

    def __init__(self, ray_test_batch: Callable, object_id_to_name: Dict = None):
        """
        Initialize the ray test helper.
        """
        self.local_transformer = LocalTransformer()
        self.ray_test_batch = ray_test_batch
        self.object_id_to_name = object_id_to_name

    def get_images_for_target(self, cam_pose: PoseStamped,
                              camera_description: 'CameraDescription',
                              camera_frame: str,
                              size: int = 256,
                              camera_min_distance: float = 0.1,
                              camera_max_distance: int = 3,
                              plot: bool = False) -> List[np.ndarray]:
        """
        Note: The returned color image is a repeated depth image in 3 channels.
        """

        # get the list of start positions of the rays.
        rays_start_positions = self.get_camera_rays_start_positions(camera_description, camera_frame, cam_pose, size,
                                                                    camera_min_distance).tolist()

        # get the list of end positions of the rays
        rays_end_positions = self.get_camera_rays_end_positions(camera_description, camera_frame, cam_pose, size,
                                                                camera_max_distance).tolist()

        # apply the ray test
        ray_test_results = self.ray_test_batch(rays_start_positions, rays_end_positions, return_distance=True)

        object_ids = [result.obj_id for result in ray_test_results]
        distances = [result.distance for result in ray_test_results]
        # construct the images/masks
        segmentation_mask = self.construct_segmentation_mask_from_ray_test_object_ids(object_ids, size)
        depth_image = self.construct_depth_image_from_ray_test_distances(distances, size) + camera_min_distance
        color_depth_image = self.construct_color_image_from_depth_image(depth_image)

        if plot:
            self.plot_segmentation_mask(segmentation_mask)
            self.plot_depth_image(depth_image)

        return [color_depth_image, depth_image, segmentation_mask]

    @staticmethod
    def construct_segmentation_mask_from_ray_test_object_ids(object_ids: List[int], size: int) -> np.ndarray:
        """
        Construct a segmentation mask from the object ids returned by the ray test.

        :param object_ids: The object ids.
        :param size: The size of the grid.
        :return: The segmentation mask.
        """
        return np.array(object_ids).squeeze(axis=1).reshape(size, size)

    @staticmethod
    def construct_depth_image_from_ray_test_distances(distances: List[float], size: int) -> np.ndarray:
        """
        Construct a depth image from the distances returned by the ray test.

        :param distances: The distances.
        :param size: The size of the grid.
        :return: The depth image.
        """
        return np.array(distances).reshape(size, size)

    @staticmethod
    def construct_color_image_from_depth_image(depth_image: np.ndarray) -> np.ndarray:
        """
        Construct a color image from the depth image.

        :param depth_image: The depth image.
        :return: The color image.
        """
        min_distance = np.min(depth_image)
        max_distance = np.max(depth_image)
        normalized_depth_image = (depth_image - min_distance) * 255 / (max_distance - min_distance)
        return np.repeat(normalized_depth_image[:, :, np.newaxis], 3, axis=2).astype(np.uint8)

    def get_camera_rays_start_positions(self, camera_description: 'CameraDescription', camera_frame: str,
                                        camera_pose: PoseStamped, size: int,
                                        camera_min_distance: float) -> np.ndarray:

        # get the start pose of the rays from the camera pose and minimum distance.
        start_pose = self.get_camera_rays_start_pose(camera_description, camera_frame, camera_pose, camera_min_distance)

        # get the list of start positions of the rays.
        return np.repeat(np.array([start_pose.position.to_list()]), size * size, axis=0)

    def get_camera_rays_start_pose(self, camera_description: 'CameraDescription', camera_frame: str, camera_pose: PoseStamped,
                                   camera_min_distance: float) -> PoseStamped:
        """
        Get the start position of the camera rays, which is the camera pose shifted by the minimum distance of the
        camera.

        :param camera_description: The camera description.
        :param camera_frame: The camera tf frame.
        :param camera_pose: The camera pose.
        :param camera_min_distance: The minimum distance from which the camera can see.
        """
        camera_transform = camera_pose.to_transform_stamped("camera_pose")
        self.local_transformer.update_transforms([camera_transform])
        camera_pose_in_camera_frame = PoseStamped(frame_id="camera_pose")
        # camera_pose_in_camera_frame = self.local_transformer.transform_pose(camera_pose, camera_frame)
        start_position = (np.array(camera_description.front_facing_axis) * camera_min_distance
                          + np.array(camera_pose_in_camera_frame.position.to_list()))
        start_pose = PoseStamped(start_position.tolist(), camera_pose_in_camera_frame.orientation.to_list(), "camera_pose")
        return self.local_transformer.transform_pose(start_pose, "map")

    def get_camera_rays_end_positions(self, camera_description: 'CameraDescription', camera_frame: str,
                                      camera_pose: PoseStamped, size: int, camera_max_distance: float = 3.0) -> np.ndarray:
        """
        Get the end positions of the camera rays.

        :param camera_description: The camera description.
        :param camera_frame: The camera frame.
        :param camera_pose: The camera pose.
        :param size: The size of the grid.
        :param camera_max_distance: The maximum distance of the camera.
        :return: The end positions of the camera rays.
        """
        rays_horizontal_angles, rays_vertical_angles = self.construct_grid_of_camera_rays_angles(camera_description,
                                                                                                 size)
        rays_end_positions = self.get_end_positions_of_rays_from_angles_and_distance(rays_vertical_angles,
                                                                                     rays_horizontal_angles,
                                                                                     camera_max_distance)
        return self.transform_points_from_camera_frame_to_world_frame(camera_pose, camera_frame, rays_end_positions)

    @staticmethod
    def transform_points_from_camera_frame_to_world_frame(camera_pose: PoseStamped, camera_frame: str,
                                                          points: np.ndarray) -> np.ndarray:
        """
        Transform points from the camera frame to the world frame.

        :param camera_pose: The camera pose.
        :param camera_frame: The camera frame.
        :param points: The points to transform.
        :return: The transformed points.
        """
        cam_to_world_transform = camera_pose.to_transform_stamped("camera_pose")
        return cam_to_world_transform.apply_transform_to_array_of_points(points)

    @staticmethod
    def get_end_positions_of_rays_from_angles_and_distance(vertical_angles: np.ndarray, horizontal_angles: np.ndarray,
                                                           distance: float) -> np.ndarray:
        """
        Get the end positions of the rays from the angles and the distance.

        :param vertical_angles: The vertical angles of the rays.
        :param horizontal_angles: The horizontal angles of the rays.
        :param distance: The distance of the rays.
        :return: The end positions of the rays.
        """
        rays_end_positions_x = distance * np.cos(vertical_angles) * np.sin(horizontal_angles)
        rays_end_positions_x = rays_end_positions_x.reshape(-1)
        rays_end_positions_z = distance * np.cos(vertical_angles) * np.cos(horizontal_angles)
        rays_end_positions_z = rays_end_positions_z.reshape(-1)
        rays_end_positions_y = distance * np.sin(vertical_angles)
        rays_end_positions_y = rays_end_positions_y.reshape(-1)
        return np.stack((rays_end_positions_x, rays_end_positions_y, rays_end_positions_z), axis=1)

    @staticmethod
    def construct_grid_of_camera_rays_angles(camera_description: 'CameraDescription',
                                             size: int) -> Tuple[np.ndarray, np.ndarray]:
        """
        Construct a 2D grid of camera rays angles.

        :param camera_description: The camera description.
        :param size: The size of the grid.
        :return: The 2D grid of the horizontal and the vertical angles of the camera rays.
        """
        # get the camera fov angles
        camera_horizontal_fov = camera_description.horizontal_angle
        camera_vertical_fov = camera_description.vertical_angle

        # construct a 2d grid of rays angles
        rays_horizontal_angles = np.linspace(-camera_horizontal_fov / 2, camera_horizontal_fov / 2, size)
        rays_horizontal_angles = np.tile(rays_horizontal_angles, (size, 1))
        rays_vertical_angles = np.linspace(-camera_vertical_fov / 2, camera_vertical_fov / 2, size)
        rays_vertical_angles = np.tile(rays_vertical_angles, (size, 1)).T
        return rays_horizontal_angles, rays_vertical_angles

    @staticmethod
    def plot_segmentation_mask(segmentation_mask,
                               object_id_to_name: Dict[int, str] = None):
        """
        Plot the segmentation mask with different colors for each object.

        :param segmentation_mask: The segmentation mask.
        :param object_id_to_name: The mapping from object id to object name.
        """
        if object_id_to_name is None:
            object_id_to_name = {uid: str(uid) for uid in np.unique(segmentation_mask)}

        # Create a custom color map
        unique_ids = np.unique(segmentation_mask)
        unique_ids = unique_ids[unique_ids != -1]  # Exclude -1 values

        # Create a color map that assigns a unique color to each ID
        colors = plt.cm.get_cmap('tab20', len(unique_ids))  # Use tab20 colormap for distinct colors
        color_dict = {uid: colors(i) for i, uid in enumerate(unique_ids)}

        # Map each ID to its corresponding color
        mask_shape = segmentation_mask.shape
        segmentation_colored = np.zeros((mask_shape[0], mask_shape[1], 3))

        for uid in unique_ids:
            segmentation_colored[segmentation_mask == uid] = color_dict[uid][:3]  # Ignore the alpha channel

        # Create a colormap for the color bar
        cmap = mcolors.ListedColormap([color_dict[uid][:3] for uid in unique_ids])
        norm = mcolors.BoundaryNorm(boundaries=np.arange(len(unique_ids) + 1) - 0.5, ncolors=len(unique_ids))

        # Plot the colored segmentation mask
        fig, ax = plt.subplots()
        _ = ax.imshow(segmentation_colored)
        ax.axis('off')  # Hide axes
        ax.set_title('Segmentation Mask with Different Colors for Each Object')

        # Create color bar
        cbar = fig.colorbar(plt.cm.ScalarMappable(norm=norm, cmap=cmap), ax=ax, ticks=np.arange(len(unique_ids)))
        cbar.ax.set_yticklabels(
            [object_id_to_name[uid] for uid in unique_ids])  # Label the color bar with object IDs
        cbar.set_label('Object Name')

        plt.show()

    @staticmethod
    def plot_depth_image(depth_image):
        # Plot the depth image
        fig, ax = plt.subplots()
        cax = ax.imshow(depth_image, cmap='viridis', vmin=0, vmax=np.max(depth_image))
        ax.axis('off')  # Hide axes
        ax.set_title('Depth Image')

        # Create color bar
        cbar = fig.colorbar(cax, ax=ax)
        cbar.set_label('Depth Value')

        plt.show()


def wxyz_to_xyzw(wxyz: List[float]) -> List[float]:
    """
    Convert a quaternion from WXYZ to XYZW format.
    """
    return [wxyz[1], wxyz[2], wxyz[3], wxyz[0]]


def xyzw_to_wxyz(xyzw: List[float]) -> List[float]:
    """
    Convert a quaternion from XYZW to WXYZ format.

    :param xyzw: The quaternion in XYZW format.
    """
    return [xyzw[3], *xyzw[:3]]


def wxyz_to_xyzw_arr(wxyz: np.ndarray) -> np.ndarray:
    """
    Convert a quaternion from WXYZ to XYZW format.

    :param wxyz: The quaternion in WXYZ format.
    """
    xyzw = np.zeros(4)
    xyzw[:3] = wxyz[1:]
    xyzw[3] = wxyz[0]
    return xyzw


def xyzw_to_wxyz_arr(xyzw: np.ndarray) -> np.ndarray:
    """
    Convert a quaternion from XYZW to WXYZ format.

    :param xyzw: The quaternion in XYZW format.
    """
    wxyz = np.zeros(4)
    wxyz[0] = xyzw[3]
    wxyz[1:] = xyzw[:3]
    return wxyz



class ClassPropertyDescriptor:
    """
    A helper that can be used to define properties of a class like the built-in ones but does not require the class
    to be instantiated.
    """

    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


def classproperty(func):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)


def is_iterable(obj: Any) -> bool:
    """
    Checks if the given object is iterable.

    :param obj: The object that should be checked
    :return: True if the object is iterable, False otherwise
    """
    try:
        iter(obj)
    except TypeError:
        return False
    return True


def lazy_product(*iterables: Iterable) -> Iterable[Tuple]:
    """
    Lazily generate the cartesian product of the iterables.

    :param iterables: Iterable of iterables to construct product for.
    :return: Iterable of tuples in the cartesian product.
    """

    consumable_iterables = [iter(iterable) for iterable in iterables]

    current_value = [next(consumable_iterable) for consumable_iterable in consumable_iterables]

    while True:
        yield tuple(current_value)

        for index in range(len(consumable_iterables) -1, -1, -1):
            current_iterable = consumable_iterables[index]
            try:
                consumable_value = next(current_iterable)
                current_value[index] = consumable_value
                break
            except StopIteration as e:
                if index == 0:
                    return
                consumable_iterables[index] = iter(iterables[index])
                current_value[index] = next(consumable_iterables[index])


# used for delayed evaluation of typing until python 3.11 becomes mainstream
from __future__ import annotations

from typing_extensions import Type, List, Tuple, Any, Dict, TYPE_CHECKING, TypeVar, Generic, Iterator, Iterable, AnyStr
from inspect import signature

from ..has_parameters import leaf_types, HasParameters
from ..language import LanguageMixin
from ..utils import is_iterable, lazy_product

T = TypeVar('T')

class PartialDesignator(LanguageMixin, Iterable[T]):
    """
    A partial designator_description is somewhat between a DesignatorDescription and a specified designator_description. Basically it is a
    partially initialized specified designator_description which can take a list of input arguments (like a DesignatorDescription)
    and generate a list of specified designators with all possible permutations of the input arguments.

    PartialDesignators are designed as generators, as such they need to be iterated over to yield the possible specified
    designators. Please also keep in mind that at the time of iteration all parameter of the specified designator_description need
    to be filled, otherwise a TypeError will be raised, see the example below for usage.

    .. code-block:: python

            # Example usage
            partial_designator = PartialDesignator(PickUpAction, milk_object_designator, arm=[Arms.RIGHT, Arms.LEFT])
            for performable in partial_designator(Grasp.FRONT):
                performable.perform()
    """
    performable: T = None
    """
    Reference to the performable class that should be initialized
    """
    args: Tuple[Any, ...] = None
    """
    Arguments that are passed to the performable
    """
    kwargs: Dict[str, Any] = None
    """
    Keyword arguments that are passed to the performable
    """

    def __init__(self, performable: T, *args, **kwargs):
        self.performable = performable
        # We use the init of the performable class since typing for the whole class messes up the signature of the class.
        # This is not optimal since "self" needs to be filtered from the parameter list but it works.
        sig = signature(self.performable.__init__)
        params_without_self = list(filter(None, [param if name != "self" else None for name, param in sig.parameters.items()]))
        sig_without_self = sig.replace(parameters=params_without_self)
        self.kwargs = dict(sig_without_self.bind_partial(*args, **kwargs).arguments)
        for key in dict(signature(self.performable).parameters).keys():
            if key not in self.kwargs.keys():
                self.kwargs[key] = None

    def __call__(self, *fargs, **fkwargs):
        """
        Creates a new PartialDesignator with the given arguments and keyword arguments added. Existing arguments will
        be prioritized over the new arguments.

        :param fargs: Additional arguments that should be added to the new PartialDesignator
        :param fkwargs: Additional keyword arguments that should be added to the new PartialDesignator
        :return: A new PartialDesignator with the given arguments and keyword arguments added
        """
        newkeywors = {k: v for k, v in self.kwargs.items() if v is not None}
        for key, value in fkwargs.items():
            if key in newkeywors.keys() and newkeywors[key] is None:
                newkeywors[key] = value
            elif key not in newkeywors.keys():
                newkeywors[key] = value
        self.kwargs.update(dict(signature(self.performable).bind_partial(*fargs, **fkwargs).arguments))
        return self

    def __iter__(self) -> Iterator[T]:
        """
        Iterates over all possible permutations of the arguments and keyword arguments and creates a new performable
        object for each permutation. In case there are conflicting parameters the args will be used over the keyword
        arguments.

        :return: A new performable object for each permutation of arguments and keyword arguments
        """
        for kwargs_combination in self.generate_permutations():
            yield self.performable(**kwargs_combination)

    def generate_permutations(self) -> Iterator[Dict[str, Any]]:
        """
        Generates the cartesian product of the given arguments. Arguments can also be a list of lists of arguments.

        :yields: A list with a possible permutation of the given arguments
        """
        iter_list = [x if is_iterable(x) and not type(x) == str else [x] for x in self.kwargs.values()]
        for combination in lazy_product(*iter_list):
            yield dict(zip(self.kwargs.keys(), combination))

    def missing_parameter(self) -> List[str]:
        """
        Returns a list of all parameters that are missing for the performable to be initialized.

        :return: A list of parameter names that are missing from the performable
        """
        missing = {k: v for k, v in self.kwargs.items() if v is None}
        return list(missing.keys())

    def resolve(self) -> T():
        """
        Returns the Designator with the first set of parameters

        :return: A fully parametrized Designator
        """
        return next(iter(self))

    def to_dict(self):
        desig = {"type": self.performable.__name__}
        desig.update(self.kwargs)
        return desig

    def flatten(self) -> List[leaf_types]:
        """
        Flattens a partial designator, very similar to HasParameters.flatten but this method can deal with parameters
        thet are None.

        :return: A list of flattened field values from the object.
        """
        result = []
        for field_name, field_type in self.performable._parameters.items():
            if self.kwargs[field_name] is not None:
                if issubclass(field_type, HasParameters):
                    sub_obj = self.kwargs[field_name]
                    result.extend(sub_obj.flatten())
                else:
                    result.append(self.kwargs[field_name])
            else:
                result.append(self.kwargs[field_name])
        return result

    def flatten_parameters(self) -> Dict[str, leaf_types]:
        """
        The flattened parameter types of the performable.

        :return: A dict with the flattened parameter types of the performable.
        """
        return self.performable.flatten_parameters()








import numpy as np
from types import MethodType
from ..tf_transformations import quaternion_from_euler

from ..pose_generator_and_validator import PoseGenerator
from ..ros import  get_ros_package_path

from ..robot_description import RobotDescription, KinematicChainDescription, EndEffectorDescription, \
    CameraDescription, RobotDescriptionManager
from ..datastructures.enums import GripperState, Arms, Grasp, TorsoState, GripperType, StaticJointState
from ..units import meter

filename = get_ros_package_path('pycram') + '/resources/robots/' + "stretch_description" + '.urdf'

stretch_description = RobotDescription("stretch_description", "base_link", "link_lift", "joint_lift", filename)

################################## Right Arm ##################################
arm_description = KinematicChainDescription("arm", "link_mast", "link_wrist_roll", stretch_description.urdf_object,
                                            arm_type=Arms.RIGHT)

arm_description.add_static_joint_states(StaticJointState.Park, {'joint_lift': 0.0,
                                                 'joint_arm_l3': 0.0,
                                                 'joint_arm_l2': 0.0,
                                                 'joint_arm_l1': 0.0,
                                                 'joint_arm_l0': 0.0,
                                                 'joint_wrist_yaw': 0.0,
                                                 'joint_wrist_pitch': 0.0,
                                                 'joint_wrist_roll': 0.0})

stretch_description.add_kinematic_chain_description(arm_description)

################################## Right Gripper ##################################
gripper_description = EndEffectorDescription("arm", "link_straight_gripper", "link_grasp_center",
                                             stretch_description.urdf_object)

gripper_description.add_static_joint_states(GripperState.OPEN, {'joint_gripper_finger_left': 0.59,
                                                                'joint_gripper_finger_right': 0.59})
gripper_description.add_static_joint_states(GripperState.CLOSE, {'joint_gripper_finger_left': 0.0,
                                                                 'joint_gripper_finger_right': 0.0})
gripper_description.end_effector_type = GripperType.PARALLEL
gripper_description.opening_distance = 0.14 * meter  # assumption
arm_description.end_effector = gripper_description

################################## Neck ##################################
neck = KinematicChainDescription("neck", "link_head", "link_head_tilt", stretch_description.urdf_object)

stretch_description.add_kinematic_chain_description(neck)

stretch_description.set_neck(yaw_joint="joint_head_pan", pitch_joint="joint_head_tilt")

################################## Torso ##################################
torso = KinematicChainDescription("torso", "link_mast", "link_lift",
                                  stretch_description.urdf_object)

torso.add_static_joint_states(TorsoState.HIGH, {"joint_lift": 1})

torso.add_static_joint_states(TorsoState.MID, {"joint_lift": 0.5})

torso.add_static_joint_states(TorsoState.LOW, {"joint_lift": 0})

stretch_description.add_kinematic_chain_description(torso)

################################## Camera ##################################
realsense_color = CameraDescription("camera_color_optical_frame", "camera_color_optical_frame", 1.322, 1.322)
realsense_depth = CameraDescription("camera_depth_optical_frame", "camera_depth_optical_frame", 1.307, 1.307)
realsense_infra1 = CameraDescription("camera_infra1_optical_frame", "camera_infra1_optical_frame", 1.307, 1.307)
realsense_infra2 = CameraDescription("camera_infra2_optical_frame", "camera_infra2_optical_frame", 1.257, 1.257)

stretch_description.add_camera_description(realsense_color)
stretch_description.add_camera_description(realsense_depth)
stretch_description.add_camera_description(realsense_infra1)
stretch_description.add_camera_description(realsense_infra2)

################################## Grasps ##################################
gripper_description.update_all_grasp_orientations([0, 0, 0, 1])

################################### Custom Orientation Generator ##############
def stretch_orientation_generator(position, origin):
    angle = np.arctan2(position[1] - origin.position.y, position[0] - origin.position.x) + np.pi + np.pi / 16
    quaternion = list(quaternion_from_euler(0, 0, angle + np.pi / 2, axes="sxyz"))
    return quaternion

################################ Load Function #################################
def load(description):
    RobotDescription.current_robot_description = description
    PoseGenerator.override_orientation_generator = stretch_orientation_generator

def unload(description):
    RobotDescription.current_robot_description = None
    PoseGenerator.override_orientation_generator = None

stretch_description.load = MethodType(load, stretch_description)
stretch_description.unload = MethodType(unload, stretch_description)

# Add to RobotDescriptionManager
rdm = RobotDescriptionManager()
rdm.register_description(stretch_description)



import unittest
from pycram.filter import Butterworth

class TestButterworthFilter(unittest.TestCase):

    def test_initialization_with_default_values(self):
        filter = Butterworth()
        self.assertEqual(filter.order, 4)
        self.assertEqual(filter.cutoff, 10)
        self.assertEqual(filter.fs, 60)

    def test_initialization_with_custom_values(self):
        filter = Butterworth(order=2, cutoff=5, fs=30)
        self.assertEqual(filter.order, 2)
        self.assertEqual(filter.cutoff, 5)
        self.assertEqual(filter.fs, 30)

    def test_filter_data_with_default_values(self):
        filter = Butterworth()
        data = [1, 2, 3, 4, 5]
        filtered_data = filter.filter(data)
        self.assertEqual(len(filtered_data), len(data))

    def test_filter_data_with_custom_values(self):
        filter = Butterworth(order=2, cutoff=5, fs=30)
        data = [1, 2, 3, 4, 5]
        filtered_data = filter.filter(data)
        self.assertEqual(len(filtered_data), len(data))

    def test_filter_empty_data(self):
        filter = Butterworth()
        data = []
        filtered_data = filter.filter(data)
        self.assertEqual(filtered_data.tolist(), data)

    def test_filter_single_value_data(self):
        filter = Butterworth()
        data = [1]
        filtered_data = filter.filter(data)
        expected_filtered_data = 0.026077721701092293  # The expected filtered value
        # self.assertAlmostEquals(filtered_data.tolist()[0], expected_filtered_data)
        self.assertAlmostEqual(filtered_data.tolist()[0], expected_filtered_data)

if __name__ == '__main__':
    unittest.main()


import math
import os
from datetime import timedelta

import yaml
from typing_extensions import Tuple, Type, Dict, Optional
from ..description import ObjectDescription
from ..object_descriptors.urdf import ObjectDescription as URDF
from ..utils import classproperty


class WorldConfig:

    """
    A class to store the configuration of the world, this can be inherited to create a new configuration class for a
    specific world (e.g. multiverse has MultiverseConfig which inherits from this class).
    """
    resources_path = os.path.realpath(os.path.join(os.path.dirname(__file__), '..', 'resources'))
    """
    Global reference for the resources path, this is used to search for the description files of the robot and
     the objects.
    """

    cache_dir_name: str = 'cached'
    """
    The name of the cache directory.
    """

    cache_dir: str = os.path.join(resources_path, cache_dir_name)
    """
    Global reference for the cache directory, this is used to cache the description files of the robot and the objects.
    """

    clear_cache_at_start: bool = True
    """
    Whether to clear the cache directory at the start.
    """

    prospection_world_prefix: str = "prospection"
    """
    The prefix for the prospection world name.
    """

    simulation_frequency: int = 240
    """
    The simulation frequency (Hz), used for calculating the equivalent real time in the simulation.
    """

    update_poses_from_sim_on_get: bool = True
    """
    Whether to update the poses from the simulator when getting the object poses.
    """

    default_description_type: Type[ObjectDescription] = URDF
    """
    The default description type for the objects.
    """

    use_physics_simulator_state: bool = False
    """
    Whether to use the physics simulator state when restoring or saving the world state.
    Currently with PyBullet, this causes a bug where ray_test does not work correctly after restoring the state using the
    simulator, so it is recommended to set this to False in PyBullet.
    """

    let_pycram_move_attached_objects: bool = True
    let_pycram_handle_spawning: bool = True
    let_pycram_handle_world_sync: bool = True
    """
    Whether to let PyCRAM handle the movement of attached objects, the spawning of objects,
     and the world synchronization.
    """

    position_tolerance: float = 1e-3
    orientation_tolerance: float = 10 * math.pi / 180
    prismatic_joint_position_tolerance: float = 1e-2
    revolute_joint_position_tolerance: float = 2 * math.pi / 180
    """
    The acceptable error for the position and orientation of an object/link, and the joint positions.
    """

    use_percentage_of_goal: bool = False
    acceptable_percentage_of_goal: float = 0.5
    """
    Whether to use a percentage of the goal as the acceptable error.
    """

    validate_goals: bool = False
    """
    Whether to validate the goals when executing them.
    """

    raise_goal_validator_error: bool = False
    """
    Whether to raise an error if the goals are not achieved.
    """

    use_giskard_monitor: bool = True
    """
    Whether to use the Giskard goal monitor when executing the goals.
    """

    allow_gripper_collision: bool = True
    """
    Whether to allow the gripper to collide with the objects when planning for the goals.
    """

    depth_images_are_in_meter: bool = False
    """
    Whether the depth images produced by :meth:`datastructures.world.World.get_images_for_target` are in meters.
    """

    max_batch_size_for_rays: Optional[int] = 16380
    """
    The maximum batch size for the rays when using the ray test batch.
    """

    execution_delay: timedelta = timedelta(seconds=0)
    """
    The delay between the execution of actions/motions to imitate real world execution time.
    """

    @classproperty
    def default_camera_config(cls) -> Dict:
        """
        The default camera configuration for the world.

        :return: The configuration for the camera as pose, yaw, tilt
        """
        if os.path.exists(os.path.join(os.path.dirname(__file__), 'camera.yaml')):
            with open(os.path.join(os.path.dirname(__file__), 'camera.yaml'), 'r') as f:
                return yaml.load(f, Loader=yaml.SafeLoader)
        return {'target_position': [1, 2, 3], 'yaw': 50, 'pitch': -35, "dist": 1.5}

    @classmethod
    def get_pose_tolerance(cls) -> Tuple[float, float]:
        return cls.position_tolerance, cls.orientation_tolerance



import time
import threading
import atexit

from sensor_msgs.msg import JointState
from std_msgs.msg import Header
from ..datastructures.world import World
from ..ros import  Time
from ..ros import  create_publisher


class JointStatePublisher:
    """
    Joint state publisher for the robot currently loaded in the World
    """
    def __init__(self, joint_state_topic="/pycram/joint_state", interval=0.1):
        """
        Robot object is from :py:attr:`~pycram.world.World.robot` and current joint states are published to
        the given joint_state_topic as a JointState message.

        :param joint_state_topic: Topic name to which the joint states should be published
        :param interval: Interval at which the joint states should be published, in seconds
        """
        self.world = World.current_world

        self.joint_state_pub = create_publisher(joint_state_topic, JointState, queue_size=10)
        self.interval = interval
        self.kill_event = threading.Event()
        self.thread = threading.Thread(target=self._publish)
        self.thread.start()

        atexit.register(self._stop_publishing)

    def _publish(self) -> None:
        """
        Publishes the current joint states of the :py:attr:`~pycram.world.World.robot` in an infinite loop.
        The joint states are published as long as the kill_event is not set by :py:meth:`~JointStatePublisher._stop_publishing`
        """
        robot = World.robot
        joint_names = [joint_name for joint_name in robot.joint_name_to_id.keys()]
        seq = 0

        while not self.kill_event.is_set():
            current_joint_states = [robot.get_joint_position(joint_name) for joint_name in joint_names]
            h = Header()
            h.stamp = Time().now()
            h.seq = seq
            h.frame_id = ""
            joint_state_msg = JointState()
            joint_state_msg.header = h
            joint_state_msg.name = joint_names
            joint_state_msg.position = current_joint_states
            # joint_state_msg.velocity = [joint_state[1] for joint_state in current_joint_states]
            self.joint_state_pub.publish(joint_state_msg)
            seq += 1
            time.sleep(self.interval)

    def _stop_publishing(self) -> None:
        """
        Sets the kill_event to terminate the publishing thread and joins the thread.
        """
        self.kill_event.set()
        self.thread.join()



from ..robot_description import RobotDescription, KinematicChainDescription, EndEffectorDescription, CameraDescription, \
    RobotDescriptionManager
from ..ros import  get_ros_package_path
from ..datastructures.enums import Arms, Grasp, GripperState, GripperType, TorsoState
from ..datastructures.dataclasses import VirtualMobileBaseJoints
from ..units import meter

filename = get_ros_package_path("pycram") + "/resources/robots/" + "kevin" + '.urdf'
kevin_description = RobotDescription("kevin", "robot_base_footprint", "robot_base_link", "robot_arm_column_joint",
                                     filename, virtual_mobile_base_joints=VirtualMobileBaseJoints())

################################## Arm ##################################
left_arm_description = KinematicChainDescription("left", "robot_arm_base_link", "robot_arm_wrist_link",
                                                 kevin_description.urdf_object, arm_type=Arms.LEFT)

left_arm_description.add_static_joint_states("park",
                                             {"robot_arm_column_joint": 0.63,
                                              "robot_arm_inner_joint": 0.03,
                                              "robot_arm_outer_joint": 4.70,
                                              "robot_arm_wrist_joint": -1.63})

kevin_description.add_kinematic_chain_description(left_arm_description)

################################## Gripper ##################################
only_gripper_description = EndEffectorDescription("gripper", "robot_arm_wrist_link", "robot_arm_tool_link",kevin_description.urdf_object)
only_gripper_description.add_static_joint_states(GripperState.OPEN, {'robot_arm_gripper_joint': 0.066,
                                              'robot_arm_gripper_mirror_joint': 0.066})
only_gripper_description.add_static_joint_states(GripperState.CLOSE, {'robot_arm_gripper_joint': 0.0,
                                               'robot_arm_gripper_mirror_joint': 0.0})
only_gripper_description.end_effector_type = GripperType.PARALLEL
only_gripper_description.opening_distance = 0.1 * meter

left_arm_description.end_effector = only_gripper_description

################################## Cameras ##################################
min_heigh = 1.2
max_height = 1.21
horizontal_angle = 0.99483  # copied from pr2
vertical_angle = 0.75049  # copied from pr2
front_facing_axis = [1, 0, 0]

top_cam = CameraDescription("top_cam", "robot_top_3d_laser_link", min_heigh,
                            max_height, horizontal_angle, vertical_angle, front_facing_axis)

kevin_description.add_camera_description(top_cam)

################################## Neck ##################################
# Everyone elese has a neck? Kevin doesn't really
# kevin_description.add_kinematic_chain("neck", "robot_front_rgbd_camera_base_link", "robot_front_rgbd_camera_link")
################################## Torso ##################################
torso = KinematicChainDescription("torso", "robot_base_link", "robot_arm_column_link",
                                  kevin_description.urdf_object)

torso.add_static_joint_states(TorsoState.HIGH, {"robot_arm_column_joint": 0.69})

torso.add_static_joint_states(TorsoState.MID, {"robot_arm_column_joint": 0.3})

torso.add_static_joint_states(TorsoState.LOW, {"robot_arm_column_joint": 0})

kevin_description.add_kinematic_chain_description(torso)
################################# Grasps ##################################
only_gripper_description.update_all_grasp_orientations([0, 0, 0, 1])

# Add to RobotDescriptionManager
rdm = RobotDescriptionManager()
rdm.register_description(kevin_description)


# used for delayed evaluation of typing until python 3.11 becomes mainstream
from __future__ import annotations

import operator

from enum import Enum
from threading import Condition, Lock
from uuid import uuid4
from typing_extensions import Any, Optional, List, Callable


class Behavior(Enum):
    """Enumeration which can be passed as argument to the pulsed method of fluents to describe how to handle missed pulses in the whenever macro.

    Fields:
    NEVER -- ignore missed pulses.
    ONCE -- if pulses were missed, execute the body once more in total.
    ALWAYS -- if pulses were missed, execute the body once more for each.
    """
    NEVER = 1
    ONCE = 2
    ALWAYS = 3


class Fluent:
    """Implementation of fluents.

    Fluents are thread-safe proxy objects which are used as variables with changing value. This allows threads to
    observe them and wait for (specific) changes.
    One can also observe fluents created by the pulsed method of a fluent. These change their value from None to True
    whenever the parent gets pulsed (or changes its value and thus gets pulsed).

    Fluents can be combined to fluent networks which allows to express complex conditions. A network updates its value
    whenever one of the fluents it is constructed from changes its value.
    The most important comparison and math operators (<, <=, ==, !=, >, >=, +, -, *, /) are overloaded to construct a
    fluent network whenever they are called with at least one fluent as parameter. In addition the comparison operators
    IS and IS_NOT as well as the logical operators AND, OR and NOT are defined as methods. This is necessary because
    these operators can't be overloaded. Fluents constructed by comparison or logical operators have the value True or
    None instead of False, so that they can be used with the wait_for method because it blocks until a value is not None.
    User defined operators can be created by passing a function as the fluents value.
    """

    def __init__(self, value: Optional[Any] = None, name: str = None):
        """Create a new fluent.

        :param value:  the value of the fluent which can also be a function to create user defined operators (default is None).
        :param name: the name of the fluent (default is a random string).
        """
        self._cv: Condition = Condition()
        self._mutex: Lock = Lock()
        self._pulses: int = 0
        self._children: List[Fluent] = []
        self._handle_missed = Behavior.NEVER
        self._value: Any = value
        self._whenever_cbs = []

        if name is not None:
            self.name: str = name
        else:
            self.name: str = str(uuid4())

    def pulsed(self, handle_missed: Behavior = 2) -> Fluent:
        """Create a fluent which changes its value from None to True whenever the parent gets pulsed.

        :param handle_missed: see the docstring of the Behavior enumeration to find out more (default is Behavior.ONCE).
        """
        fluent = Fluent()

        def value():
            if fluent._pulses != 0:
                return True
            else:
                return None

        fluent.set_value(value)
        fluent._handle_missed = handle_missed
        self.add_child(fluent)
        return fluent

    def pulse(self) -> None:
        """Pulse a fluent without changing its value."""
        for child in self._children:
            with child._mutex:
                child._pulses += 1

            child.pulse()
        for callback in self._whenever_cbs:
            callback(self.get_value())
            with self._mutex:
                self._pulses -= 1
            if self._handle_missed == Behavior.NEVER:
                with self._mutex:
                    self._pulses = 0
            elif self._handle_missed == Behavior.ONCE:
                callback(self.get_value())
                with self._mutex:
                    self._pulses = 0
            elif self._handle_missed == Behavior.ALWAYS:
                with self._mutex:
                    for i in range(self._pulses):
                        callback(self.get_value())
        with self._cv:
            self._cv.notify()

    def whenever(self, callback: Callable) -> None:
        """
        Registers a callback which is called everytime this Fluent is pulsed. The callback should be a Callable. When
        the callback is called it gets the current value of this Fluent as an argument.

        :param callback: The callback which should be called when pulsed as a Callable.
        """
        self._whenever_cbs.append(callback)

    def add_child(self, child: Fluent) -> None:
        """Add a child to the fluent which gets pulsed whenever this fluent gets pulsed, too.

        :param child: the child to add.
        """
        self._children.append(child)

    def get_value(self) -> Any:
        """Return the value of the fluent."""
        with self._mutex:
            if callable(self._value):
                return self._value()

            return self._value

    def set_value(self, value: Any) -> None:
        """Change the value of the fluent.
        Changing the value will also pulse the fluent.

        :param value: the new value of the fluent.
        """
        with self._mutex:
            self._value = value

        self.pulse()

    def wait_for(self, timeout: Optional[float] = None):
        """Block the current thread if the value of the fluent is None, until it is not None or until it timed out.

        If the fluent was created by the pulsed method of a fluent, the method blocks until the parent gets pulsed.

        The return value is the last return value of the predicate (value is not None) and will evaluate to False if the method timed out.

        :param timeout: the maximum time to wait (default is None).
        """
        with self._cv:
            return self._cv.wait_for(lambda: self.get_value() is not None, timeout)

    def _compare(self, operator, other: Fluent) -> Fluent:
        """This is a helper method for internal usage only.

        Create a fluent which value is a function returning True or None depending on the given comparison operator applied to the operands.

        :param operator: the comparison operator to apply.
        :param other: the other operand.
        """
        fluent = Fluent()

        if type(other) == Fluent:
            def value():
                if operator(self.get_value(), other.get_value()):
                    return True
                else:
                    return None

            other.add_child(fluent)
        else:
            def value():
                if operator(self.get_value(), other):
                    return True
                else:
                    return None

        self.add_child(fluent)
        fluent.set_value(value)
        return fluent

    def __lt__(self, other: Fluent) -> Fluent:
        """Overload the < comparsion operator.

        :param other: -- the other operand.
        """
        return self._compare(operator.lt, other)

    def __leq__(self, other: Fluent) -> Fluent:
        """Overload the <= comparsion operator.

        :param other: the other operand.
        """
        return self._compare(operator.leq, other)

    def __eq__(self, other: Fluent) -> Fluent:
        """Overload the == comparsion operator.

        :param other: the other operand.
        """
        return self._compare(operator.eq, other)

    def __ne__(self, other: Fluent) -> Fluent:
        """Overload the != comparsion operator.

        :param other: the other operand.
        """
        return self._compare(operator.ne, other)

    def IS(self, other: Fluent) -> Fluent:
        """Create a fluent which value is True if the value of its parent is the value of the given operand, None otherwise.

        :param other: -- the other operand which can also be a fluent.
        """
        return self._compare(operator.is_, other)

    def IS_NOT(self, other: Fluent) -> Fluent:
        """Create a fluent which value is True if the value of its parent is not the value of the given operand, None otherwise.

        :param other: -- the other operand which can also be a fluent.
        """
        return self._compare(operator.is_not, other)

    def __gt__(self, other: Fluent) -> Fluent:
        """Overload the > comparison operator.

        :param other: the other operand.
        """
        return self._compare(operator.gt, other)

    def __geq__(self, other: Fluent) -> Fluent:
        """Overload the >= comparison operator.

        :param other: the other operand.
        """
        return self._compare(operator.geq, other)

    def _math(self, operator: Callable, operand: Fluent, other: Fluent) -> Fluent:
        """This is a helper method for internal usage only.

        Create a fluent which value is a function returning the value of the given math operator applied to the operands.

        :param operator: the math operator to apply.
        :param operand: the first operand.
        :param other: the other operand.
        """
        fluent = Fluent()

        if type(operand) == Fluent:
            if type(other) == Fluent:
                value = lambda: operator(operand.get_value(), other.get_value())
                other.add_child(fluent)
            else:
                value = lambda: operator(operand.get_value(), other)

            operand.add_child(fluent)
        else:
            value = lambda: operator(operand, other.get_value())
            other.add_child(fluent)

        fluent.set_value(value)
        return fluent

    def __add__(self, other: Fluent) -> Fluent:
        """Overload the + math operator.

        :parm other: the other operand.
        """
        return self._math(operator.add, self, other)

    def __radd__(self, other: Fluent) -> Fluent:
        """Overload the + math operator with the first operand not being a fluent.

        :param other: the other operand.
        """
        return self._math(operator.add, other, self)

    def __sub__(self, other: Fluent) -> Fluent:
        """Overload the - math operator.

        :param other: the other operand.
        """
        return self._math(operator.sub, self, other)

    def __rsub__(self, other: Fluent) -> Fluent:
        """Overload the - math operator with the first operand not being a fluent.

        :param other: the other operand.
        """
        return self._math(operator.sub, other, self)

    def __mul__(self, other: Fluent) -> Fluent:
        """Overload the * math operator.

        :param other: the other operand.
        """
        return self._math(operator.mul, self, other)

    def __rmul__(self, other: Fluent) -> Fluent:
        """Overload the * math operator with the first operand not being a fluent.

        :param other: the other operand.
        """
        return self._math(operator.mul, other, self)

    def __truediv__(self, other: Fluent) -> Fluent:
        """Overload the / math operator.

        :param other: the other operand.
        """
        return self._math(operator.truediv, self, other)

    def __rtruediv__(self, other) -> Fluent:
        """Overload the / math operator with the first operand not being a fluent.

        :param other: the other operand.
        """
        return self._math(operator.truediv, other, self)

    def AND(self, other: Fluent) -> Fluent:
        """Create a fluent which value is True if both, the value of its parent and the other operand express True, None otherwise.

        :param other: the other operand which can also be a fluent.
        """
        fluent = Fluent()

        if type(other) == Fluent:
            def value():
                if self.get_value() and other.get_value():
                    return True
                else:
                    return None

            other.add_child(fluent)
        else:
            def value():
                if self.get_value() and other:
                    return True
                else:
                    return None

        self.add_child(fluent)
        fluent.set_value(value)
        return fluent

    def OR(self, other: Fluent) -> Fluent:
        """Create a fluent which value is True if either the value of its parent or the other operand express True, None otherwise.

        :param other: the other operand which can also be a fluent.
        """
        fluent = Fluent()

        if type(other) == Fluent:
            def value():
                if self.get_value() or other.get_value():
                    return True
                else:
                    return None

            other.add_child(fluent)
        else:
            def value():
                if self.get_value() or other:
                    return True
                else:
                    return None

        self.add_child(fluent)
        fluent.set_value(value)
        return fluent

    def NOT(self) -> Fluent:
        """Create a fluent which value is True if the value of its parent expresses False, None otherwise."""
        def value():
            if not self.get_value():
                return True
            else:
                return None

        fluent = Fluent(value)
        self.add_child(fluent)
        return fluent



def create_subscriber(topic, msg_type, callback, queue_size=10):
    """
    Create a ROS subscriber.

    :param topic: The name of the topic to subscribe to.
    :param msg_type: The type of message to subscribe to.
    :param callback: The callback function to call when a message is received.
    :param queue_size: The size of the message queue.
    :return: A ROS subscriber object.
    """
    # This is a placeholder implementation. In a real implementation, you would use
    # the appropriate ROS API to create the subscriber.
    raise NotImplementedError("ROS is not available in this environment. ")


import numpy as np

from .. import tf_transformations
from typing_extensions import List, Union, Tuple, Dict

from ..ros import get_node_names
from ..ros import  Duration, ServiceException
from ..ros import  loginfo_once, logerr
from ..ros import  get_service_proxy, wait_for_service
from ..datastructures.world import World, UseProspectionWorld
from ..world_concepts.world_object import Object
from ..utils import _apply_ik
from ..local_transformer import LocalTransformer
from ..datastructures.pose import PoseStamped
from ..robot_description import RobotDescription
from ..failures import IKError
from ..external_interfaces.giskard import projection_cartesian_goal, allow_gripper_collision
from .pinocchio_ik import compute_ik
from ..datastructures.enums import Arms

try:
    from moveit_msgs.msg import PositionIKRequest
    from moveit_msgs.msg import RobotState
    from moveit_msgs.srv import GetPositionIK
    from sensor_msgs.msg import JointState
except ImportError:
    pass



def _make_request_msg(root_link: str, tip_link: str, target_pose: PoseStamped, robot_object: Object,
                      joints: List[str]) -> 'PositionIKRequest':
    """
    Generates an ik request message for the kdl_ik_service. The message is
    of the type moveit_msgs/PositionIKRequest and contains all information
    contained in the parameter.

    :param root_link: The first link of the chain of joints to be altered
    :param tip_link: The last link of the chain of joints to be altered
    :param target_pose: Target pose for which the message should be created
    :param robot_object: The robot for which the ik should be generated
    :param joints: A list of joint names between the root_link and tip_link that should be altered.
    :return: A moveit_msgs/PositionIKRequest message containing all the information from the parameter
    """
    local_transformer = LocalTransformer()
    target_pose = local_transformer.transform_pose(target_pose, robot_object.get_link_tf_frame(root_link))

    robot_state = RobotState()
    joint_state = JointState()
    joint_state.name = joints
    joint_state.position = [robot_object.get_joint_position(joint) for joint in joints]
    # joint_state.velocity = [0.0 for x in range(len(joints))]
    # joint_state.effort = [0.0 for x in range(len(joints))]
    robot_state.joint_state = joint_state

    msg_request = PositionIKRequest()
    # msg_request.group_name = "arm"
    msg_request.ik_link_name = tip_link
    msg_request.pose_stamped = target_pose
    msg_request.avoid_collisions = False
    msg_request.robot_state = robot_state
    msg_request.timeout = Duration(1000)
    # msg_request.attempts = 1000

    return msg_request


def call_ik(root_link: str, tip_link: str, target_pose: PoseStamped, robot_object: Object, joints: List[str]) -> List[float]:
    """
   Sends a request to the kdl_ik_service and returns the solution.
   Note that the robot in robot_object should be identical to the robot description
   uploaded to the parameter server. Furthermore, note that the root_link and
   tip_link are the links attached to the first and last joints in the joints list.

   :param root_link: The first link of the chain of joints to be altered
   :param tip_link: The last link in the chain of joints to be altered
   :param target_pose: The target pose in frame of root link second is the orientation as quaternion in world coordinate frame
   :param robot_object: The robot object for which the ik solution should be generated
   :param joints: A list of joint name that should be altered
   :return: The solution that was generated as a list of joint values corresponding to the order of joints given
   """
    if RobotDescription.current_robot_description.name == "pr2":
        ik_service = "/pr2_right_arm_kinematics/get_ik" if "r_wrist" in tip_link else "/pr2_left_arm_kinematics/get_ik"
    else:
        ik_service = "/kdl_ik_service/get_ik"

    loginfo_once(f"Waiting for IK service: {ik_service}")
    wait_for_service(ik_service)

    req = _make_request_msg(root_link, tip_link, target_pose, robot_object, joints)
    req.pose_stamped.header.frame_id = root_link
    ik = get_service_proxy(ik_service, GetPositionIK)
    try:
        resp = ik(req)
    except ServiceException as e:
        if RobotDescription.current_robot_description.name == "pr2":
            raise IKError(target_pose, root_link, tip_link)
        else:
            raise e

    if resp.error_code.val == -31:
        raise IKError(target_pose, root_link, tip_link)

    return resp.solution.joint_state.position


def try_to_reach_with_grasp(pose_or_object: Union[PoseStamped, Object],
                            prospection_robot: Object, gripper_name: str,
                            grasp_quaternion: List[float]) -> Union[PoseStamped, None]:
    """
    Checks if the robot can reach a given position with a specific grasp orientation.
    To determine this the inverse kinematics are calculated and applied.

    :param pose_or_object: The position and rotation or Object for which reachability should be checked or an Object
    :param prospection_robot: The robot that should reach for the position
    :param gripper_name: The name of the end effector
    :param grasp_quaternion: The orientation of the grasp
    """

    input_pose = pose_or_object.get_pose() if isinstance(pose_or_object, Object) else pose_or_object

    target_pose = apply_grasp_orientation_to_pose(grasp_quaternion, input_pose)

    return try_to_reach(target_pose, prospection_robot, gripper_name)


def apply_grasp_orientation_to_pose(grasp_orientation: List[float], pose: PoseStamped) -> PoseStamped:
    """
    Applies the orientation of a grasp to a given pose. This is done by using the grasp orientation
    of the given grasp and applying it to the given pose.

    :param grasp_orientation: The orientation of the grasp
    :param pose: The pose to which the grasp orientation should be applied
    """
    local_transformer = LocalTransformer()
    target_map = local_transformer.transform_pose(pose, "map")
    target_map.orientation.x = grasp_orientation[0]
    target_map.orientation.y = grasp_orientation[1]
    target_map.orientation.z = grasp_orientation[2]
    target_map.orientation.w = grasp_orientation[3]
    return target_map


def try_to_reach(pose_or_object: Union[PoseStamped, Object], prospection_robot: Object,
                 gripper_name: str) -> Union[PoseStamped, None]:
    """
    Tries to reach a given position with a given robot. This is done by calculating the inverse kinematics.

    :param pose_or_object: The position and rotation or Object for which reachability should be checked.
    :param prospection_robot: The robot that should be used to check for reachability, should be the one in the prospection world
    :param gripper_name: Name of the gripper tool frame
    :return: The pose at which the robot should stand or None if the target is not reachable
    """
    input_pose = pose_or_object.get_pose() if isinstance(pose_or_object, Object) else pose_or_object

    arm_chain = list(filter(lambda chain: chain.get_tool_frame() == gripper_name, RobotDescription.current_robot_description.get_manipulator_chains()))[0]

    joints = arm_chain.joints

    try:
        inv = request_ik(input_pose, prospection_robot, joints, gripper_name)
    except IKError as e:
        logerr(f"Pose is not reachable: {e}")
        return None
    _apply_ik(prospection_robot, inv)

    return input_pose


def request_ik(target_pose: PoseStamped, robot: Object, joints: List[str], gripper: str) -> Tuple[PoseStamped, Dict[str, float]]:
    """
    Top-level method to request ik solution for a given pose. This method will check if the giskard node is running
    and if so will call the giskard service. If the giskard node is not running the kdl_ik_service will be called.

    :param target_pose: Pose of the end-effector for which an ik solution should be found
    :param robot: The robot object which should be used
    :param joints: A list of joints that should be used in computation, this is only relevant for the kdl_ik_service
    :param gripper: Name of the tool frame which should grasp, this should be at the end of the given joint chain
    :return: A Pose at which the robt should stand as well as a dictionary of joint values
    """
    if "/giskard" not in get_node_names():
        return robot.pose, request_pinocchio_ik(target_pose, robot, gripper, joints)
        # return robot.pose, request_kdl_ik(target_pose, robot, joints, gripper)
    return request_giskard_ik(target_pose, robot, gripper)


def request_kdl_ik(target_pose: PoseStamped, robot: Object, joints: List[str], gripper: str) -> Dict[str, float]:
    """
    Top-level method to request ik solution for a given pose. Before calling the ik service the links directly before
    and after the joint chain will be queried and the target_pose will be transformed into the frame of the root_link.
    Afterward, the offset between the tip_link and end effector will be calculated and taken into account. Lastly the
    ik service is called and the result returned

    :param target_pose: Pose for which an ik solution should be found
    :param robot: Robot object which should be used
    :param joints: List of joints that should be used in computation
    :param gripper: Name of the gripper which should grasp, this should be at the end of the given joint chain
    :return: A list of joint values
    """
    local_transformer = LocalTransformer()
    base_link = RobotDescription.current_robot_description.get_parent(joints[0])
    # Get link after last joint in chain
    end_effector = RobotDescription.current_robot_description.get_child(joints[-1])

    target_torso = local_transformer.transform_pose(target_pose, robot.get_link_tf_frame(base_link))

    wrist_tool_frame_offset = robot.get_transform_between_links(end_effector, gripper)
    target_diff = target_torso.to_transform_stamped("target").inverse_times(wrist_tool_frame_offset).to_pose_stamped()

    inv = call_ik(base_link, end_effector, target_diff, robot, joints)

    return dict(zip(joints, inv))


def request_giskard_ik(target_pose: PoseStamped, robot: Object, gripper: str) -> Tuple[PoseStamped, Dict[str, float]]:
    """
    Calls giskard in projection mode and queries the ik solution for a full body ik solution. This method will
    try to drive the robot directly to a pose from which the target_pose is reachable for the end effector. If there
    are obstacles in the way this method will fail. In this case please use the GiskardLocation designator.

    :param target_pose: Pose at which the end effector should be moved.
    :param robot: Robot object which should be used.
    :param gripper: Name of the tool frame which should grasp, this should be at the end of the given joint chain.
    :return: A list of joint values.
    """
    loginfo_once(f"Using Giskard for full body IK")
    local_transformer = LocalTransformer()
    target_map = local_transformer.transform_pose(target_pose, "map")

    allow_gripper_collision(Arms.BOTH)
    result = projection_cartesian_goal(target_map, gripper, "map")
    last_point = result.trajectory.points[-1]
    joint_names = result.trajectory.joint_names

    joint_states = dict(zip(joint_names, last_point.positions))
    prospection_robot = World.current_world.get_prospection_object_for_object(robot)

    orientation = list(tf_transformations.quaternion_from_euler(0, 0, joint_states["brumbrum_yaw"], axes="sxyz"))
    pose = PoseStamped.from_list([joint_states["brumbrum_x"], joint_states["brumbrum_y"], 0], orientation)

    robot_joint_states = {}
    for joint_name, state in joint_states.items():
        if joint_name in robot.joints.keys():
            robot_joint_states[joint_name] = state

    with UseProspectionWorld():
        prospection_robot.set_multiple_joint_positions(robot_joint_states)
        prospection_robot.set_pose(pose)

        tip_pose = prospection_robot.get_link_pose(gripper)
        dist = tip_pose.position.euclidean_distance(target_map.position)

        if dist > 0.01:
            raise IKError(target_pose, "map", gripper)
        return pose, robot_joint_states

def request_pinocchio_ik(target_pose: PoseStamped, robot: Object, target_link: str, joints: List[str]) -> Dict[str, float]:
    """
    Calls the pinocchio ik solver to calculate the ik solution for a given target link and pose.

    :param target_link: The target link for which the ik solution should be calculated
    :param target_pose: The target pose for which the ik solution should be calculated
    :param robot: The robot object for which the ik solution should be calculated
    :param joints: The joints that should be used in the calculation
    :return: A dictionary containing the joint names and joint values
    """
    lt = LocalTransformer()
    target_pose = lt.transform_pose(target_pose, "map")
    target_pose = lt.transform_pose(target_pose, robot.tf_frame)

    # Get link after last joint in chain
    wrist_link = RobotDescription.current_robot_description.get_child(joints[-1])

    wrist_tool_frame_offset = robot.get_transform_between_links(wrist_link, target_link)
    target_diff = target_pose.to_transform_stamped("target").inverse_times(wrist_tool_frame_offset).to_pose_stamped()
    target_diff.round()


    res = compute_ik(wrist_link, target_diff, robot)

    return res


from .default_process_modules import *
from ..datastructures.enums import GripperState, Arms, ExecutionType
from ..datastructures.world import World
from ..designators.motion_designator import MoveGripperMotion
from ..process_module import ProcessModule, ProcessModuleManager
from ..ros import Duration, create_action_client
from ..ros import  loginfo, logwarn, logdebug

try:
    from pr2_controllers_msgs.msg import Pr2GripperCommandGoal, Pr2GripperCommandAction, Pr2
except ImportError:
    pass

try:
    from ..worlds.multiverse import Multiverse
except ImportError:
    Multiverse = type(None)

try:
    from control_msgs.msg import GripperCommandGoal, GripperCommandAction
except ImportError:
    if Multiverse is not None:
        logwarn("Import for control_msgs for gripper in Multiverse failed")

try:
    from ..worlds import Multiverse
except ImportError:
    Multiverse = type(None)

try:
    from pr2_controllers_msgs.msg import Pr2GripperCommandGoal, Pr2GripperCommandAction, Pr2
except ImportError:
    logdebug("Pr2GripperCommandGoal not found")

class Pr2MoveGripperMultiverse(ProcessModule):
    """
    Opens or closes the gripper of the real PR2, gripper uses an action server for this instead of giskard
    """

    def _execute(self, designator: MoveGripperMotion):
        def activate_callback():
            loginfo("Started gripper Movement")

        def done_callback(state, result):
            loginfo(f"Reached goal {designator.motion}: {result.reached_goal}")

        def feedback_callback(msg):
            loginfo(f"Gripper Action Feedback: {msg}")

        goal = GripperCommandGoal()
        goal.command.position = 0.0 if designator.motion == GripperState.CLOSE else 0.4
        goal.command.max_effort = 50.0
        if designator.gripper == "right":
            controller_topic = "/real/pr2/right_gripper_controller/gripper_cmd"
        else:
            controller_topic = "/real/pr2/left_gripper_controller/gripper_cmd"
        client = create_action_client(controller_topic, GripperCommandAction)
        loginfo("Waiting for action server")
        client.wait_for_server()
        client.send_goal(goal, active_cb=activate_callback, done_cb=done_callback, feedback_cb=feedback_callback)
        wait = client.wait_for_result(Duration(5))
        # client.cancel_all_goals()

class Pr2MoveGripperReal(ProcessModule):
    """
    Opens or closes the gripper of the real PR2, gripper uses an action server for this instead of giskard
    """

    def _execute(self, designator: MoveGripperMotion):
        def activate_callback():
            loginfo("Started gripper Movement")

        def done_callback(state, result):
            loginfo(f"Reached goal {designator.motion}")

        def feedback_callback(msg):
            pass

        goal = Pr2GripperCommandGoal()
        goal.command.position = 0.0 if designator.motion == GripperState.CLOSE else 0.1
        goal.command.max_effort = 50.0
        if designator.gripper == Arms.RIGHT:
            controller_topic = "r_gripper_controller/gripper_action"
        else:
            controller_topic = "l_gripper_controller/gripper_action"
        client = create_action_client(controller_topic, Pr2GripperCommandAction)
        loginfo("Waiting for action server")
        client.wait_for_server()
        client.send_goal(goal, active_cb=activate_callback, done_cb=done_callback, feedback_cb=feedback_callback)
        wait = client.wait_for_result()

class Pr2Manager(DefaultManager):
    def __init__(self):
        super().__init__()
        self.robot_name = "pr2"

    def move_gripper(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveGripper(self._move_gripper_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            if (isinstance(World.current_world, Multiverse) and
                    World.current_world.conf.use_multiverse_process_modules):
                return Pr2MoveGripperMultiverse(self._move_gripper_lock)
            else:
                return Pr2MoveGripperReal(self._move_gripper_lock)

# Initialize the PR2 manager and register it with the process module manager
Pr2Manager()